import paraview
from paraview import servermanager

from paraview.util import proxy as proxy_util

from paraview.simple.session import GetActiveView
from paraview.simple.view import GetViews, GetRenderViews

# -----------------------------------------------------------------------------


def RenameLayout(newName, proxy=None):
    """Renames the given layout. If the given proxy is not registered
    in the 'layouts' group this method will have no effect.

    :param newName: The new name of the layout proxy
    :type newName: str
    :param layout: The layout proxy to rename. It must be a member
        of the 'layouts' groupd. Optional, defaults to renaming the active layout."""
    if not proxy:
        proxy = GetLayout()
    proxy_util.rename(proxy, "layouts", newName)


def CreateLayout(name=None):
    """Create a new layout with no active view.

    :param name: The name of the layout. Optional, defaults to an automatically
        created name."""
    layout = servermanager.misc.ViewLayout(registrationGroup="layouts")
    if name:
        RenameLayout(name, layout)
    return layout


# -----------------------------------------------------------------------------


def RemoveLayout(proxy=None):
    """Remove the provided layout. If none is provided, remove the layout
    containing the active view. If it is the only layout it will create a new
    one with the same name as the removed one.

    :param proxy: The layout proxy to remove. Optional, defaults to the
        layout containing the active view.
    :type proxy: Layout proxy or `None`."""
    pxm = servermanager.ProxyManager()
    if not proxy:
        proxy = GetLayout()
    name = pxm.GetProxyName("layouts", proxy)
    pxm.UnRegisterProxy("layouts", name, proxy)
    if len(GetLayouts()) == 0:
        CreateLayout(name)


# -----------------------------------------------------------------------------


def GetLayouts():
    """Returns all the layout proxies.

    :return: A list of all the layouts.
    :rtype: list of layout proxies"""
    return servermanager.ProxyManager().GetProxiesInGroup("layouts")


# -----------------------------------------------------------------------------


def GetLayout(view=None):
    """Return the layout containing the given view, if any.

    :param view: A view in the layout to be removed. Optional, defaults to the
        active view.
    :return: The layout containing the view
    :rtype: :class:`paraview.servermanager.ViewLayout`"""
    if not view:
        view = GetActiveView()
    if not view:
        raise RuntimeError("No active view was found.")
    lproxy = servermanager.vtkSMViewLayoutProxy.FindLayout(view.SMProxy)
    return servermanager._getPyProxy(lproxy)


def GetLayoutByName(name):
    """Return the first layout with the given name, if any.

    :param name: Name of the layout
    :type name: str
    :return: The named layout if it exists
    :rtype: Layout proxy or `None`"""
    layouts = GetLayouts()
    for key in layouts.keys():
        if key[0] == name:
            return layouts.get(key)
    return None


def GetViewsInLayout(layout=None):
    """Returns a list of views in the given layout.

    :param layout: Layout whose views should be returned. Optional, defaults to the
        layout for the active view, if possible.
    :type layout: Layout proxy.
    :return: List of views in the layout.
    :rtype: list"""
    layout = layout if layout else GetLayout()
    if not layout:
        raise RuntimeError(
            "Layout couldn't be determined. Please specify a valid layout."
        )
    views = GetViews()
    return [x for x in views if layout.GetViewLocation(x) != -1]


def AssignViewToLayout(view=None, layout=None, hint=0):
    """Assigns the view provided to the layout provided. If no layout exists,
    then a new layout will be created.

    It is an error to assign the same view to multiple layouts.

    :param view: The view to assign to the layout. Optional, defaults to the active view.
    :type view: View proxy.
    :param layout: If layout is `None`, then either the active layout or an
                   existing layout on the same server will be used.
    :type layout: Layout proxy.
    :return: Returns `True` on success.
    :rtype: bool

    """
    view = view if view else GetActiveView()
    if not view:
        raise RuntimeError("No active view was found.")

    layout = layout if layout else GetLayout()
    controller = servermanager.ParaViewPipelineController()
    return controller.AssignViewToLayout(view, layout, hint)


# -----------------------------------------------------------------------------


def RemoveViewsAndLayouts():
    """
    Removes all existing views and layouts.

    """
    pxm = servermanager.ProxyManager()
    layouts = pxm.GetProxiesInGroup("layouts")

    for view in GetRenderViews():
        proxy_util.unregister(view)

    # Can not use regular delete for layouts
    for name, id in layouts:
        proxy = layouts[(name, id)]
        pxm.UnRegisterProxy("layouts", name, proxy)


def EqualizeViewsHorizontally(layout=None):
    """Equalizes horizontal view sizes in the provided layout.

    :param layout: Layout the layout contain the views to equalize. Optional, defaults
                   to the active layout.
    :type layout: Layout proxy.
    """
    layout = layout if layout else GetLayout()
    layout.SMProxy.EqualizeViews(
        paraview.modules.vtkRemotingViews.vtkSMViewLayoutProxy.HORIZONTAL
    )


def EqualizeViewsVertically(layout=None):
    """Equalizes vertical view sizes in the provided layout.

    :param layout: Layout the layout contain the views to equalize. Optional, defaults
                   to the active layout.
    :type layout: Layout proxy.
    """
    layout = layout if layout else GetLayout()
    layout.SMProxy.EqualizeViews(
        paraview.modules.vtkRemotingViews.vtkSMViewLayoutProxy.VERTICAL
    )


def EqualizeViewsBoth(layout=None):
    """Equalizes the vertical and horizontal view sizes in the provided layout.

    :param layout: Layout the layout contain the views to equalize. Optional, defaults
                   to the active layout.
    :type layout: Layout proxy.
    """
    layout = layout if layout else GetLayout()
    layout.SMProxy.EqualizeViews()



from paraview import servermanager


def GetSettingsProxy(type):
    """Given a string giving the type of settings to access, returns the
    proxy for those settings.

    :param type: The type of settings to access. Valid types may be found
        by calling :func:`GetAllSettings()`.
    :type type: str"""
    pxm = servermanager.ProxyManager()
    proxy = pxm.GetProxy("settings", type)
    return proxy


def GetAllSettings():
    """Get a list of strings that return valid settings proxy types for the
    :func:GetSettingsProxy() function.

    :return: List of valid settings proxy names
    :rtype: list of str"""
    settingsList = []
    pxm = servermanager.ProxyManager()
    pdm = pxm.GetProxyDefinitionManager()
    iter = pdm.NewSingleGroupIterator("settings")
    iter.GoToFirstItem()

    while not iter.IsDoneWithTraversal():
        settingsList.append(iter.GetProxyName())
        iter.GoToNextItem()
    return settingsList



import sys
import os
import paraview
import paraview.simple as pvsimple

can_savecinema = True
try:
    from vtkmodules.numpy_interface.algorithms import *
    from mpi4py import MPI
except:
    can_savecinema = False

paraview.options.batch = True  # this may not be necessary
paraview.simple._DisableFirstRenderCameraReset()


def CreateTimeCompartments(globalController, timeCompartmentSize):
    if globalController.GetNumberOfProcesses() == 1:
        print('single process')
        return
    elif globalController.GetNumberOfProcesses() % timeCompartmentSize != 0:
        print('number of processes must be an integer multiple of time compartment size')
        return
    elif timeCompartmentSize == globalController.GetNumberOfProcesses():
        return globalController

    gid = globalController.GetLocalProcessId()
    timeCompartmentGroupId = int(gid / timeCompartmentSize)
    newController = globalController.PartitionController(timeCompartmentGroupId, gid % timeCompartmentSize)
    # must unregister if the reference count is greater than 1
    if newController.GetReferenceCount() > 1:
        newController.UnRegister(None)

    # print (gid, ' of global comm is ', newController.GetLocalProcessId())
    globalController.SetGlobalController(newController)
    return newController


def CheckReader(reader):
    if hasattr(reader, "FileName") == False:
        print("ERROR: Don't know how to set file name for ", reader.SMProxy.GetXMLName())
        sys.exit(-1)

    if hasattr(reader, "TimestepValues") == False:
        print("ERROR: ", reader.SMProxy.GetXMLName(), " doesn't have time information")
        sys.exit(-1)


def CreateControllers(timeCompartmentSize):
    pm = paraview.servermanager.vtkProcessModule.GetProcessModule()
    globalController = pm.GetGlobalController()
    if timeCompartmentSize > globalController.GetNumberOfProcesses():
        timeCompartmentSize = globalController.GetNumberOfProcesses()

    temporalController = CreateTimeCompartments(globalController, timeCompartmentSize)
    return globalController, temporalController, timeCompartmentSize


def WriteImages(currentTimeStep, currentTime, views):
    cinemaLines = []
    cnt = 0
    for view in views:
        filename = view.tpFileName.replace("%t", str(currentTimeStep))
        view.ViewTime = currentTime
        cinemaLines.append(str(currentTime) + "," + view.GetXMLName() + "_" + str(cnt) + "," + filename + "\n")
        pvsimple.SaveScreenshot(filename, view, magnification=view.tpMagnification)
        cnt = cnt + 1
    return cinemaLines


def WriteFiles(currentTimeStep, currentTime, writers):
    cinemaLines = []
    cnt = 0
    for writer in writers:
        originalfilename = writer.FileName
        fname = originalfilename.replace("%t", str(currentTimeStep))
        writer.FileName = fname
        cinemaLines.append(str(currentTime) + "," + writer.GetXMLName() + "_" + str(cnt) + "," + fname + "\n")
        writer.UpdatePipeline(currentTime)
        writer.FileName = originalfilename
        cnt = cnt + 1
    return cinemaLines


def IterateOverTimeSteps(globalController, timeCompartmentSize, timeSteps, writers, views, make_cinema_table=False):
    if make_cinema_table and not can_savecinema:
        print("WARNING: Can not save cinema table because MPI4PY is not available.")
        make_cinema_table = False

    numProcs = globalController.GetNumberOfProcesses()
    numTimeCompartments = numProcs // timeCompartmentSize
    tpp = len(timeSteps) / numTimeCompartments
    remainder = len(timeSteps) % numTimeCompartments
    timeCompartmentIndex = int(globalController.GetLocalProcessId() / timeCompartmentSize)
    myStartTimeStep = tpp * timeCompartmentIndex
    myEndTimeStep = myStartTimeStep + tpp
    if timeCompartmentIndex < remainder:
        myStartTimeStep = myStartTimeStep + timeCompartmentIndex
        myEndTimeStep = myStartTimeStep + tpp + 1
    else:
        myStartTimeStep = myStartTimeStep + remainder
        myEndTimeStep = myStartTimeStep + tpp

    myStartTimeStep = int(myStartTimeStep)
    myEndTimeStep = int(myEndTimeStep)
    cinemaLines = []
    for currentTimeStep in range(myStartTimeStep, myEndTimeStep):
        # print (globalController.GetLocalProcessId(), " is working on ", currentTimeStep)
        ret = WriteImages(currentTimeStep, timeSteps[currentTimeStep], views)
        if ret:
            cinemaLines.extend(ret)
        ret = WriteFiles(currentTimeStep, timeSteps[currentTimeStep], writers)
        if ret:
            cinemaLines.extend(ret)

    if make_cinema_table:
        # gather the file list from each time compartment to the root and save it
        myGID = globalController.GetLocalProcessId()
        myLID = int(myGID) % int(timeCompartmentSize)
        mystring = ''
        # only one node per time compartment should participate
        if myLID == 0:
            mystring = ''.join(x for x in cinemaLines)
        mylen = len(mystring)
        result = ''
        comm = vtkMPI4PyCommunicator.ConvertToPython(globalController.GetCommunicator())
        gathered_lines = comm.gather(mystring, root=0)
        if myGID == 0:
            # root writes the file, prepending header
            f = open("data.csv", "w")
            f.write("timestep,producer,FILE\n")
            for x in gathered_lines:
                if x:
                    f.write(x)
            f.close()


def CreateReader(ctor, fileInfo, **kwargs):
    "Creates a reader, checks if it can be used, and sets the filenames"
    import glob
    files = glob.glob(fileInfo)
    files.sort()  # assume there is a logical ordering of the filenames that corresponds to time ordering
    reader = paraview.simple.OpenDataFile(files)
    CheckReader(reader)
    if kwargs:
        pvsimple.SetProperties(reader, **kwargs)
    return reader


def CreateWriter(ctor, filename, tp_writers):
    writer = ctor()
    return RegisterWriter(writer, filename, tp_writers)


def RegisterWriter(writer, filename, tp_writers):
    writer.FileName = filename
    tp_writers.append(writer)
    return writer


def CreateView(proxy_ctor, filename, magnification, width, height, tp_views):
    view = proxy_ctor()
    return RegisterView(view, filename, magnification, width, height, tp_views)


def RegisterView(view, filename='filename_{time}.vti', magnification=1.0, width=1024, height=1024, tp_views=[]):
    view.add_attribute("tpFileName", filename)
    view.add_attribute("tpMagnification", magnification)
    tp_views.append(view)
    view.ViewSize = [width, height]
    return view



# SPDX-FileCopyrightText: Copyright (c) Kitware Inc.
# SPDX-License-Identifier: BSD-3-Clause

r"""collaboration is a helper module useful for handling synchronisation
in multi-client configuration as well as providing other collaboration
related methods

A simple example:
  from paraview import collaboration

  collaboration.processServerEvents()
"""

import paraview, re, types, sys, vtk
from paraview import servermanager


def processServerEvents():
    """Update the local state based on the notifications received from the server
    that have been generated by other clients."""
    if servermanager.ActiveConnection:
        session = servermanager.ActiveConnection.Session
        if session.IsMultiClients() and session.IsNotBusy():
            while servermanager.vtkProcessModule.GetProcessModule().GetNetworkAccessManager().ProcessEvents(100):
                pass



import http.server
import socketserver
from urllib.parse import urlparse
from urllib.parse import parse_qs
from os import chdir
from os import path
from os.path import relpath
from os import getcwd
from os import access
from os import R_OK
import pathlib
import json
from .. import cdb

#
# global variables - can't seem to add an instance variable to the
# subclass of SimpleHTTPRequestHandler
#
CinemaInstallPath = "CINEMAJUNK"
ServerInstallPath = "CINEMAJUNK"

def set_install_path():
    global CinemaInstallPath
    global ServerInstallPath

    ServerInstallPath = str(pathlib.Path(__file__).parent.absolute())
    CinemaInstallPath = str(pathlib.Path(__file__).parent.absolute())
    # edit the path to get the correct installation path
    CinemaInstallPath = CinemaInstallPath.strip("/server")
    CinemaInstallPath = "/" + CinemaInstallPath + "/viewers"

def verify_cinema_databases( runpath, databases ):
    result = False

    curdir = getcwd()
    chdir(runpath)
    for db in databases:
        if not path.isdir(db):
            print("")
            print("ERROR: Cinema server is looking for the database:") 
            print("           '{}\',".format(db)) 
            print("       but it does not exist relative to the server runpath:")
            print("           '{}\',".format(runpath)) 
            print("")
            result = False
            break;
        else:
            result = True

    chdir(curdir)
    return result 

#
# CinemaSimpleReqestHandler
#
# Processes GET requests to find viewers and databases
#
class CinemaSimpleRequestHandler(http.server.SimpleHTTPRequestHandler):

    @property
    def silent(self):
        return self._silent

    @silent.setter
    def silent(self, value):
        self._silent = value

    @property
    def rundir(self):
        return self._rundir

    @rundir.setter
    def rundir(self, value):
        self._rundir = value

    @property
    def databases(self):
        return self._databases

    @databases.setter
    def databases(self, value):
        self._databases = value

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        self._verbose = value

    @property
    def assetname(self):
        return self._assetname

    @assetname.setter
    def assetname(self, value):
        self._assetname = value

    @property
    def viewer(self):
        return self._viewer

    @viewer.setter
    def viewer(self, value):
        self._viewer = value

    def translate_path(self, path ):
        # print("translate_path: {}".format(path))
        return path

    def log(self, message):
        if self.verbose:
            print(message)

    def log_message( self, format, *args ):
        pass

    def do_GET(self):
        global CinemaInstallPath
        global ServerInstallPath

        self.log(" ")
        self.log("PATH     : {}".format(self.path))

        # the request is for a viewer
        if self.path == "/":
            for db in self.databases:
                if not path.isdir(db):
                    self.log("ERROR")
                    self.path = ServerInstallPath + "/error_no-database.html"
                    return http.server.SimpleHTTPRequestHandler.do_GET(self)

            if self.viewer == "explorer": 
                # handle a request for the Cinema:Explorer viewer
                self.log("EXPLORER")
                self.path = CinemaInstallPath + "/cinema_explorer.html"
                self.log(self.path)
                return http.server.SimpleHTTPRequestHandler.do_GET(self)

            elif self.viewer == "view": 
                # handle a request for the Cinema:View viewer
                self.log("VIEW")
                self.path = CinemaInstallPath + "/cinema_view.html"
                return http.server.SimpleHTTPRequestHandler.do_GET(self)

            if self.viewer == "test": 
                # handle a request for the Cinema:Explorer viewer
                self.log("TEST")
                self.path = ServerInstallPath + "/cinema_test.html"
                return http.server.SimpleHTTPRequestHandler.do_GET(self)

            else:
                self.log("VIEWER: {}".format(self.viewer))


        if self.path.endswith("cinema_attributes.json"):
            # this is a request to the server for attributes
            self.log("ATTRIBUTE REQUEST") 

            if (not self.assetname == None):
                json_string = "{{\"assetname\" : \"{}\"}}".format(self.assetname)
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.send_header("Content-length", len(json_string))
                self.end_headers()
                self.wfile.write(str.encode(json_string))
            return

        if self.path.endswith("databases.json"):
            self.log("DATABASES   : {}".format(self.path))
            json_string = self.get_database_json()

            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.send_header("Content-length", len(json_string))
            self.end_headers()
            self.wfile.write(json_string.encode())
            return


        if self.path.startswith("/cinema/"):
            # handle a requests for sub components of the viewers 
            # NOTE: fragile - requires 'cinema' path be unique

            self.log("CINEMA   : {}".format(self.path))
            self.path = CinemaInstallPath + self.path 
            self.log("         : {}".format(self.path))
            return http.server.SimpleHTTPRequestHandler.do_GET(self)

        else:
            # everything else
            self.log("NORMAL   : {}".format(self.path))
            self.path = relpath(self.path.strip("/"), getcwd())
            # self.log("UPDATED  : {}".format(self.path))
            return http.server.SimpleHTTPRequestHandler.do_GET(self)

    def get_database_json( self ):
        dbj = [] 
        if   self.viewer == "explorer": 
            for db in self.databases:
                dbj.append({    "name": db,
                                "directory": db 
                           })
        elif self.viewer == "view":
            dbs = []
            for db in self.databases:
                dbs.append( {
                                "name": db,
                                "location": db
                            }
                          )
            dbj.append({    
                        "database_name": db,
                        "datasets": dbs,
                       })
        elif self.viewer == "simple": 
            for db in self.databases:
                dbj.append(db)

        return json.dumps(dbj, indent=4)

def run_cinema_server( viewer, rundir, databases, port, assetname="FILE", verbose=False, silent=False):
    localhost = "http://127.0.0.1"

    if (verbose):
        print("Running cinema server:")
        print("    rundir   : {}".format(rundir))
        print("    viewer   : {}".format(viewer))
        print("    databases: {}".format(databases))
        print("    port     : {}".format(port))
        print("    assetname: {}".format(assetname))
        print("")

    expanded_rundir = path.expanduser(rundir)
    fullpath = path.abspath(expanded_rundir)
    if verify_cinema_databases(fullpath, databases) :
        chdir(fullpath)
        set_install_path()
        cin_handler = CinemaSimpleRequestHandler
        cin_handler.verbose   = verbose
        cin_handler.viewer    = viewer
        cin_handler.assetname = assetname
        cin_handler.databases = databases
        socketserver.TCPServer.allow_reuse_address = True
        with socketserver.TCPServer(("", port), cin_handler) as httpd:
            urlstring = "{}:{}".format(localhost, port)
            if not silent:
                print(urlstring)
            try:
                httpd.serve_forever()
            except( KeyboardInterrupt, Exception ) as e:
                if not silent:
                    print(str(e))




import paraview
import paraview._backwardscompatibilityhelper

from paraview import servermanager
from paraview.util import proxy as proxy_util

# ==============================================================================
# Internal helpers
# ==============================================================================


def _create_func(key, module, skipRegistration=False):
    """Internal function."""

    from paraview.simple.session import active_objects

    def CreateObject(*input, **params):
        """
        This function creates a new proxy. For pipeline objects that accept inputs,
        all non-keyword arguments are assumed to be inputs. All keyword arguments are
        assumed to be property-value pairs and are passed to the new proxy.

        """

        registrationName = None
        for nameParam in ["registrationName", "guiName"]:
            if nameParam in params:
                registrationName = params[nameParam]
                del params[nameParam]

        # Create a controller instance.
        controller = servermanager.ParaViewPipelineController()

        from paraview.catalyst.detail import IsInsituInput, CreateProducer

        if IsInsituInput(registrationName):
            # This is a catalyst input, replace with a trivial producer
            px = CreateProducer(registrationName)
        else:
            # Instantiate the actual object from the given module.
            px = paraview._backwardscompatibilityhelper.GetProxy(
                module, key, no_update=True
            )

        # preinitialize the proxy.
        controller.PreInitializeProxy(px)

        # Make sure non-keyword arguments are valid
        for inp in input:
            if inp != None and not isinstance(inp, servermanager.Proxy):
                if px.GetProperty("Input") != None:
                    raise RuntimeError("Expecting a proxy as input.")
                else:
                    raise RuntimeError(
                        "This function does not accept non-keyword arguments."
                    )

        # Assign inputs
        inputName = servermanager.vtkSMCoreUtilities.GetInputPropertyName(px.SMProxy, 0)

        if px.GetProperty(inputName) != None:
            if len(input) > 0:
                px.SetPropertyWithName(inputName, input)
            else:
                # If no input is specified, try the active pipeline object
                if (
                    px.GetProperty(inputName).GetRepeatable()
                    and active_objects.get_selected_sources()
                ):
                    px.SetPropertyWithName(
                        inputName, active_objects.get_selected_sources()
                    )
                elif active_objects.source:
                    px.SetPropertyWithName(inputName, active_objects.source)
        else:
            if len(input) > 0:
                raise RuntimeError("This function does not expect an input.")

        # Pass all the named arguments as property,value pairs
        proxy_util.set(px, **params)

        # post initialize
        controller.PostInitializeProxy(px)

        if isinstance(px, servermanager.MultiplexerSourceProxy):
            px.UpdateDynamicProperties()

        if not skipRegistration:
            # Register the proxy with the proxy manager (assuming we are only using
            # these functions for pipeline proxies or animation proxies.
            if isinstance(px, servermanager.SourceProxy):
                controller.RegisterPipelineProxy(px, registrationName)
            elif px.GetXMLGroup() == "animation":
                controller.RegisterAnimationProxy(px)
        return px

    # add special tag to detect these functions in _remove_functions
    CreateObject.__paraview_create_object_tag = True
    CreateObject.__paraview_create_object_key = key
    CreateObject.__paraview_create_object_module = module
    return CreateObject


# -----------------------------------------------------------------------------


def _listProperties(create_func):
    """Internal function that, given a proxy creation function, e.g.,
    paraview.simple.Sphere, returns the list of named properties that can
    be set during construction."""
    key = create_func.__paraview_create_object_key
    module = create_func.__paraview_create_object_module
    prototype_func = _create_func(key, module, skipRegistration=True)

    # Find the prototype by XML label name
    xml_definition = module._findProxy(name=key)
    xml_group = xml_definition["group"]
    xml_name = xml_definition["key"]
    prototype = servermanager.ProxyManager().GetPrototypeProxy(xml_group, xml_name)

    # Iterate over properties and add them to the list
    property_iter = prototype.NewPropertyIterator()
    property_iter.UnRegister(None)
    property_names = []
    while not property_iter.IsAtEnd():
        label = property_iter.GetPropertyLabel()
        if label is None:
            label = property_iter.GetKey()
        property_names.append(paraview.make_name_valid(label))
        property_iter.Next()

    return property_names


# -----------------------------------------------------------------------------


def _get_proxymodules_to_import(connection):
    """
    used in _add_functions, _get_generated_proxies, and _remove_functions to get
    modules to import proxies from.
    """
    if connection and connection.ProxiesNS:
        modules = connection.ProxiesNS
        return [modules.filters, modules.sources, modules.writers, modules.animation]
    else:
        return []


# -----------------------------------------------------------------------------


def _get_generated_proxies():
    proxies = []
    for m in _get_proxymodules_to_import(servermanager.ActiveConnection):
        for key in dir(m):
            proxies.append(key)
    return proxies


# -----------------------------------------------------------------------------


def _find_writer(filename):
    "Internal function."
    extension = None
    parts = filename.split(".")
    if len(parts) > 1:
        extension = parts[-1]
    else:
        raise RuntimeError("Filename has no extension, please specify a write")

    if extension == "png":
        return "vtkPNGWriter"
    elif extension == "bmp":
        return "vtkBMPWriter"
    elif extension == "ppm":
        return "vtkPNMWriter"
    elif extension == "tif" or extension == "tiff":
        return "vtkTIFFWriter"
    elif extension == "jpg" or extension == "jpeg":
        return "vtkJPEGWriter"
    else:
        raise RuntimeError("Cannot infer filetype from extension:", extension)


# -----------------------------------------------------------------------------


class _funcs_internals:
    "Internal class."
    first_render = True


# -----------------------------------------------------------------------------


def ListProperties(proxyOrCreateFunction):
    """Given a proxy or a proxy creation function, e.g. paraview.simple.Sphere,
    returns the list of properties for the proxy that would be created.

    :param proxyOrCreateFunction: Proxy or proxy creation function whose property
        names are desired.
    :type proxyOrCreateFunction: Proxy or proxy creation function
    :return: List of property names
    :rtype: List of str"""

    try:
        return _listProperties(proxyOrCreateFunction)
    except:
        pass

    try:
        return proxyOrCreateFunction.ListProperties()
    except:
        pass

    return None


# -----------------------------------------------------------------------------


def _DisableFirstRenderCameraReset():
    """Normally a ResetCamera is called automatically when Render is called for
    the first time after importing the `paraview.simple` module. Calling this
    function disables this first render camera reset."""
    _funcs_internals.first_render = False


# -----------------------------------------------------------------------------


def _create_doc(new, old):
    "Internal function."
    res = new + "\n"
    ts = []
    strpd = old.split("\n")
    for s in strpd:
        ts.append(s.lstrip())
    res += " ".join(ts)
    res += "\n"
    return res


# -----------------------------------------------------------------------------


def _func_name_valid(name):
    "Internal function."
    for c in name:
        if c == "(" or c == ")":
            return False

    return True


# -----------------------------------------------------------------------------


def _get_function_arguments(function):
    """The a list of named arguments for a function.
    Used for autocomplete in PythonShell in ParaView GUI and in pvpython"""
    import inspect

    if not inspect.isfunction(function) and not inspect.ismethod(function):
        raise TypeError(f"input argument: {function} is not a function")
    result = []
    if hasattr(function, "__paraview_create_object_tag") and getattr(
        function, "__paraview_create_object_tag"
    ):
        result = ListProperties(function)
    else:
        for _, parameter in inspect.signature(function).parameters.items():
            # skip *args and **kwargs
            if (
                parameter.kind != inspect.Parameter.VAR_POSITIONAL
                and parameter.kind != inspect.Parameter.VAR_KEYWORD
            ):
                result.append(parameter.name)
    return result


def _add_functions(g):
    if not servermanager.ActiveConnection:
        return []

    added_entries = set()
    activeModule = servermanager.ActiveConnection.ProxiesNS

    # add deprecated proxies first, so we create the new function and documentation.
    modules = paraview._backwardscompatibilityhelper.get_deprecated_proxies(
        activeModule
    )
    for proxyModule in modules:
        for deprecatedProxyPair in modules[proxyModule]:
            oldProxyLabel = deprecatedProxyPair[0]
            newProxyLabel = deprecatedProxyPair[1]
            if not newProxyLabel in g and _func_name_valid(newProxyLabel):
                f = _create_func(deprecatedProxyPair, proxyModule)
                f.__doc__ = "{} is a deprecated proxy. It will be automatically replaced by {}".format(
                    oldProxyLabel, newProxyLabel
                )
                f.__qualname__ = oldProxyLabel
                f.__name__ = oldProxyLabel
                g[oldProxyLabel] = f
                added_entries.add(oldProxyLabel)

    for m in _get_proxymodules_to_import(servermanager.ActiveConnection):
        # Skip registering proxies in certain modules (currently only writers)
        skipRegistration = m is activeModule.writers
        for key in dir(m):
            if key not in g and _func_name_valid(key):
                # print "add %s function" % key
                f = _create_func(key, m, skipRegistration)
                f.__qualname__ = key
                f.__name__ = key
                f.__doc__ = _create_doc(m.getDocumentation(key), f.__doc__)
                g[key] = f
                added_entries.add(key)

    return list(added_entries)


def _remove_functions(g):
    to_remove = [
        item[0]
        for item in g.items()
        if hasattr(item[1], "__paraview_create_object_tag")
    ]
    for key in to_remove:
        del g[key]
        # paraview.print_info("remove %s", key)


def _switchToActiveConnectionCallback(caller, event):
    """Callback called when the active session/connection changes in the
    ServerManager. We update the Python state to reflect the change."""
    if servermanager:
        session = servermanager.vtkSMProxyManager.GetProxyManager().GetActiveSession()
        connection = servermanager.GetConnectionFromSession(session)
        SetActiveConnection(connection)


def _initializeSession(connection):
    """Internal method used to initialize a session. Users don't need to
    call this directly. Whenever a new session is created this method is called
    by API in this module."""
    if not connection:
        raise RuntimeError("'connection' cannot be empty.")
    controller = servermanager.ParaViewPipelineController()
    controller.InitializeSession(connection.Session)


class _active_session_observer:
    def __init__(self):
        pxm = servermanager.vtkSMProxyManager.GetProxyManager()
        self.ObserverTag = pxm.AddObserver(
            pxm.ActiveSessionChanged, _switchToActiveConnectionCallback
        )

    def __del__(self):
        if servermanager:
            if servermanager.vtkSMProxyManager:
                servermanager.vtkSMProxyManager.GetProxyManager().RemoveObserver(
                    self.ObserverTag
                )


# -----------------------------------------------------------------------------


class _active_objects(object):
    """This class manages the active objects (source and view). The active
    objects are shared between Python and the user interface. This class
    is for internal use. Use the :ref:`SetActiveSource`,
    :ref:`GetActiveSource`, :ref:`SetActiveView`, and :ref:`GetActiveView`
    methods for setting and getting active objects."""

    def __get_selection_model(self, name, session=None):
        "Internal method."
        if session and session != servermanager.ActiveConnection.Session:
            raise RuntimeError(
                "Try to set an active object with invalid active connection."
            )
        pxm = servermanager.ProxyManager(session)
        model = pxm.GetSelectionModel(name)
        if not model:
            model = servermanager.vtkSMProxySelectionModel()
            pxm.RegisterSelectionModel(name, model)
        return model

    def set_view(self, view):
        "Sets the active view."
        active_view_model = self.__get_selection_model("ActiveView")
        if view:
            active_view_model = self.__get_selection_model(
                "ActiveView", view.GetSession()
            )
            active_view_model.SetCurrentProxy(
                view.SMProxy, active_view_model.CLEAR_AND_SELECT
            )
        else:
            active_view_model = self.__get_selection_model("ActiveView")
            active_view_model.SetCurrentProxy(None, active_view_model.CLEAR_AND_SELECT)

    def get_view(self):
        "Returns the active view."
        return servermanager._getPyProxy(
            self.__get_selection_model("ActiveView").GetCurrentProxy()
        )

    def set_source(self, source):
        "Sets the active source."
        active_sources_model = self.__get_selection_model("ActiveSources")
        if source:
            # 3 == CLEAR_AND_SELECT
            active_sources_model = self.__get_selection_model(
                "ActiveSources", source.GetSession()
            )
            active_sources_model.SetCurrentProxy(
                source.SMProxy, active_sources_model.CLEAR_AND_SELECT
            )
        else:
            active_sources_model = self.__get_selection_model("ActiveSources")
            active_sources_model.SetCurrentProxy(
                None, active_sources_model.CLEAR_AND_SELECT
            )

    def __convert_proxy(self, px):
        "Internal method."
        if not px:
            return None
        elif px.IsA("vtkSMOutputPort"):
            return servermanager.OutputPort(
                servermanager._getPyProxy(px.GetSourceProxy()), px.GetPortIndex()
            )
        else:
            return servermanager._getPyProxy(px)

    def get_source(self):
        "Returns the active source."
        return self.__convert_proxy(
            self.__get_selection_model("ActiveSources").GetCurrentProxy()
        )

    def get_selected_sources(self):
        "Returns the set of sources selected in the pipeline browser."
        model = self.__get_selection_model("ActiveSources")
        proxies = []
        for i in range(model.GetNumberOfSelectedProxies()):
            proxies.append(self.__convert_proxy(model.GetSelectedProxy(i)))
        return proxies

    view = property(get_view, set_view)
    source = property(get_source, set_source)


# ==============================================================================
# Active Source / View / Camera / AnimationScene
# ==============================================================================


def GetActiveView():
    """Returns the active view.

    :return: Active view.
    :rtype: View proxy."""
    return active_objects.view


# -----------------------------------------------------------------------------


def SetActiveView(view):
    """Sets the active view.

    :param view: The view to make active.
    :type view: View proxy."""
    active_objects.view = view


# -----------------------------------------------------------------------------


def GetActiveSource():
    """Returns the active source.

    :return: Active pipeline source.
    :rtype: Source proxy"""
    return active_objects.source


# -----------------------------------------------------------------------------


def SetActiveSource(source):
    """Sets the active source.

    :param source: The source
    :type source: Source proxy"""
    active_objects.source = source


# -----------------------------------------------------------------------------


def GetActiveCamera():
    """Returns the active camera for the active view.

    :return: The active camera
    :rtype: `vtkCamera`"""
    return GetActiveView().GetActiveCamera()


# ==============================================================================
# Client/Server Connection methods
# ==============================================================================


def Disconnect(ns=None, force=True):
    """Disconnect from the currently connected server and free the active
    session. Does not shut down the client application where the call is executed.

    :param ns: Namespace in which ParaView functions were created. Optional, defaults
        to the namespace returned by `globals()`
    :type ns: Dict or `None`
    :param force: Force disconnection in a simultaneous connection. Optional, defaults
        to forcing disconnection.
    :type force: bool
    """
    if not ns:
        ns = globals()

    supports_simutaneous_connections = (
        servermanager.vtkProcessModule.GetProcessModule().GetMultipleSessionsSupport()
    )
    if not force and supports_simutaneous_connections:
        # This is an internal Disconnect request that doesn't need to happen in
        # multi-server setup. Ignore it.
        return
    if servermanager.ActiveConnection:
        _remove_functions(ns)
        servermanager.Disconnect()
        import gc

        gc.collect()


# -----------------------------------------------------------------------------


def Connect(
    ds_host=None, ds_port=11111, rs_host=None, rs_port=11111, timeout=60, ns=None
):
    """Creates a connection to a server.

    :param ds_host: Data server hostname. Needed to set up a client/data server/render server connection.
    :type ds_host: string
    :param ds_port: Data server port to listen on.
    :type ds_port: int
    :param rs_host: Render server hostname. Needed to set up a client/data server/render server connection.
    :type rs_host: string
    :param rs_port: Render server port to listen on.
    :type rs_port: int
    :param timeout: Time in seconds at which the connection is abandoned if not made.
    :type timeout: int

    Example usage connecting to a host named "amber"::

        Connect("amber") # Connect to a single server at default port
        Connect("amber", 12345) # Connect to a single server at port 12345
        Connect("amber", 11111, "vis_cluster", 11111) # connect to data server, render server pair
        Connect("amber", timeout=30) # Connect to a single server at default port with a 30s timeout instead of default 60s
        Connect("amber", timeout=-1) # Connect to a single server at default port with no timeout instead of default 60s
        Connect("amber", timeout=0)  # Connect to a single server at default port without retrying instead of retrying for
        the default 60s

    """
    from paraview.simple import _extend_simple

    if ns is None:
        ns = globals()
    Disconnect(ns, False)
    connection = servermanager.Connect(ds_host, ds_port, rs_host, rs_port, timeout)
    if not (connection is None):
        _initializeSession(connection)
        _extend_simple(ns)
    return connection


# -----------------------------------------------------------------------------


def ReverseConnect(port=11111, ns=None):
    """Create a reverse connection to a server.  First, disconnects from any servers,
    then listens on port and waits for an incoming connection from the server.

    :param port: The port to listen on for incoming connections.
    :type port: int
    :return: Connection object
    :rtype:
    """
    from paraview.simple import _extend_simple

    if ns is None:
        ns = globals()
    Disconnect(ns, False)
    connection = servermanager.ReverseConnect(port)
    _initializeSession(connection)
    _extend_simple(ns)
    return connection


# -----------------------------------------------------------------------------


def ResetSession():
    """Reset the session to its initial state. All pipeline readers, sources,
    filters extractors, and representations are removed."""
    connection = servermanager.ResetSession()
    _initializeSession(connection)
    return connection


# ==============================================================================
# Multi-servers
# ==============================================================================


def SetActiveConnection(connection=None, ns=None):
    """Set the active connection. If the process was run without multi-server
    enabled and this method is called with a non-None argument while an
    ActiveConnection is present, it will raise a RuntimeError.

    :param connection: If provided, changes the current connection.
    :type connection: Connection object.
    :param ns: Namespace in which functions from the old Connection are removed
        and functions in the new Connection are added.
    :raises RuntimeError: If called when ParaView is not running in multi-server
        mode, a `RuntimeError` will be raised.
    """
    from paraview.simple import _extend_simple

    if not ns:
        ns = globals()
    if servermanager.ActiveConnection != connection:
        _remove_functions(ns)
        servermanager.SetActiveConnection(connection)
        _extend_simple(ns)


# -----------------------------------------------------------------------------


def IsFirstRender():
    return _funcs_internals.first_render


# -----------------------------------------------------------------------------


def ResetFirstRender():
    _funcs_internals.first_render = False


# -----------------------------------------------------------------------------
# Initialize module
# -----------------------------------------------------------------------------

if not paraview.options.satellite:
    active_session_observer = _active_session_observer()
    active_objects = _active_objects()



# --------------------------------------------------------------------------
# Handle BooleanDomain
# --------------------------------------------------------------------------
def booleanDomainDecorator(props, xmlProps, uiProps, domain):
    uiProps["widget"] = "checkbox"
    uiProps["type"] = "int"
    return True


# --------------------------------------------------------------------------
# Handle EnumerationDomain
# --------------------------------------------------------------------------
def enumerationDomainDecorator(props, xmlProps, uiProps, domain):
    uiProps["widget"] = "list-1"
    uiProps["type"] = "int"

    valMap = {}
    for entryNum in range(domain.GetNumberOfEntries()):
        valMap[domain.GetEntryText(entryNum)] = domain.GetEntryValue(entryNum)

    uiProps["values"] = valMap

    return True


# --------------------------------------------------------------------------
# Handle StringListDomain
# --------------------------------------------------------------------------
def stringListDomainDecorator(props, xmlProps, uiProps, domain):
    uiProps["widget"] = "list-1"
    uiProps["type"] = "str"

    valuesList = []
    for idx in range(domain.GetNumberOfStrings()):
        valuesList.append(domain.GetString(idx))

    uiProps["values"] = valuesList

    return True


# --------------------------------------------------------------------------
# Handle TreeDomain
# --------------------------------------------------------------------------
def numpad(numelts):
    return len(str(int(numelts)))


def treeDomainDecorator(props, xmlProps, uiProps, domain):
    uiProps["type"] = "int"
    uiProps["widget"] = "list-n"

    values = []

    index = 0
    stack = []
    stack.append([domain.GetInformation(), ["Multi-block Dataset"], -1])

    while len(stack) > 0:
        me = stack.pop()

        dataInformation = me[0]
        accumulatedName = me[1]
        parentIndex = me[2]

        values.append([accumulatedName, index, parentIndex])

        infoName = None

        if dataInformation:
            infoName = dataInformation.GetCompositeDataClassName()

        if infoName:
            # May have children
            info = dataInformation.GetCompositeDataInformation()
            numChildren = info.GetNumberOfChildren()
            if numChildren > 0:
                for i in range(numChildren - 1, -1, -1):
                    child = info.GetDataInformation(i)
                    childName = info.GetName(i)
                    stack.append([child, accumulatedName + [childName], index])

        index += 1

    # valMap = { 'Block 1': 2, 'Block 2': 3 }

    valMap = {}
    blockList = [None] * len(values)
    pads = numpad(len(values))
    for value in values:
        names = value[0]
        blockIdx = value[1]
        parentIdx = value[2]
        padding = "-" * (len(names) - 1)
        valuesEntryName = "%s %s %s" % (str(blockIdx).zfill(pads), padding, names[-1])
        valMap[valuesEntryName] = blockIdx
        blockList[blockIdx] = {"label": names[-1], "parent": parentIdx, "id": blockIdx}

    uiProps["values"] = valMap
    uiProps["blocks"] = blockList
    uiProps["sort"] = True

    return True


# --------------------------------------------------------------------------
# Handle ArrayListDomain
# --------------------------------------------------------------------------
def arrayListDomainDecorator(props, xmlProps, uiProps, domain):
    uiProps["type"] = "str"
    uiProps["widget"] = "list-1"

    fieldMap = {"0": "POINTS", "1": "CELLS"}
    valuesMap = {}
    valuesList = []

    for arnum in range(domain.GetNumberOfStrings()):
        stringName = domain.GetString(arnum)
        try:
            assocVar = domain.GetFieldAssociation(arnum)
            valuesMap[stringName] = [fieldMap[str(assocVar)], stringName]
        except:
            valuesList.append(stringName)

    if len(valuesList) > 0:
        uiProps["values"] = valuesList
    else:
        uiProps["values"] = valuesMap

    return True


# --------------------------------------------------------------------------
# Handle ArraySelectionDomain
# --------------------------------------------------------------------------
def arraySelectionDomainDecorator(props, xmlProps, uiProps, domain):
    uiProps["type"] = "str"
    uiProps["widget"] = "list-n"

    valuesList = []

    for arnum in range(domain.GetNumberOfStrings()):
        valuesList.append(domain.GetString(arnum))

    uiProps["values"] = valuesList

    return True


# --------------------------------------------------------------------------
# Handle IntRangeDomain, DoubleRangeDomain, and ArrayRangeDomain
# --------------------------------------------------------------------------
def numberRangeDomainDecorator(props, xmlProps, uiProps, domain):
    uiProps["widget"] = "textfield"

    domainClass = domain.GetClassName()

    if domainClass.rfind("Int") >= 0:
        uiProps["type"] = "int"
    else:
        uiProps["type"] = "float"

    ranges = []

    idx = 0
    while domain.GetMinimumExists(idx) == 1 and domain.GetMaximumExists(idx) == 1:
        ranges.append({"min": domain.GetMinimum(idx), "max": domain.GetMaximum(idx)})
        idx += 1

    uiProps["range"] = ranges

    if idx == 0:
        return False
    return True


# --------------------------------------------------------------------------
# Handle ProxyListDomain
# --------------------------------------------------------------------------
def proxyListDomainDecorator(props, xmlProps, uiProps, domain):
    uiProps["widget"] = "list-1"
    uiProps["size"] = 1
    uiProps["type"] = "proxy"
    values = {}

    for i in range(domain.GetNumberOfProxies()):
        nextProxy = domain.GetProxy(i)
        values[nextProxy.GetXMLLabel()] = nextProxy.GetGlobalIDAsString()

    uiProps["values"] = values
    return True


# --------------------------------------------------------------------------
# Decorate for PropertyWidgetDecorator hint with type="ClipScalarDecorator"
# --------------------------------------------------------------------------
# FIXME: Remove this when GenericDecorator goes into master
def clipScalarDecorator(prop, uiElt, hint):
    proxy = None
    try:
        proxy = prop.SMProperty.GetParent()
    except:
        try:
            proxy = prop.GetParent()
        except:
            print("ERROR: unable to get proxy for property " + prop.Name)
            return
    excludeAttr = hint.GetAttribute("exclude")
    if not excludeAttr or excludeAttr != "1":
        depends = proxy.GetGlobalIDAsString() + ":ClipFunction:Scalar:1"
    else:
        depends = proxy.GetGlobalIDAsString() + ":ClipFunction:Scalar:0"
    uiElt["depends"] = depends


# --------------------------------------------------------------------------
# Decorate for PropertyWidgetDecorator hint with type="GenericDecorator"
# --------------------------------------------------------------------------
def genericDecorator(prop, uiElt, hint):
    proxy = None
    try:
        proxy = prop.SMProperty.GetParent()
    except:
        try:
            proxy = prop.GetParent()
        except:
            print("ERROR: unable to get proxy for property " + prop.Name)
            return

    mode = hint.GetAttribute("mode")

    # For now we just handle visbility mode
    if mode == "visibility":
        propName = hint.GetAttribute("property")
        value = hint.GetAttribute("value")
        inverse = hint.GetAttribute("inverse")
        visibility = "1"

        if not inverse or inverse != "1":
            visibility = "1"
        else:
            visibility = "0"

        uiElt["depends"] = "%s:%s:%s:%s" % (
            proxy.GetGlobalIDAsString(),
            propName,
            value,
            visibility,
        )


# --------------------------------------------------------------------------
# Decorate for Widget hint with type="multi_line"
# --------------------------------------------------------------------------
def multiLineDecorator(prop, uiElt, hint):
    uiElt["widget"] = "textarea"


# --------------------------------------------------------------------------
# Decorate for proxyEditorPropertyWidget hint
# --------------------------------------------------------------------------
def proxyEditorPropertyWidgetDecorator(prop, uiElt, hint):
    propName = hint.GetAttribute("property")
    if propName:
        uiElt["visibilityProperty"] = propName



import h5py
import numpy
import glob
import cinemasci.cdb
import os.path

class ascent():

    def __init__(self):
        self.outcdb    = None
        self.basename  = None 
        self.dirname   = None 
        self.cycles    = []
        self.minmax    = {}
        self.variables = [] 


    #
    # convert an ascent hdf5 data set of float images to a cinema database
    #
    # this function looks in the "indir" directory for files matching the pattern:
    #    <something>.cycle_<0 padded number>
    #
    def convert(self, indir, outcdb):
        self.indir  = indir
        self.outcdb = outcdb

        # initialize the cinema database 
        cdb = cinemasci.new("cdb", {"path": self.outcdb})
        cdb.initialize()

        self.__get_cycles()
        self.__get_variable_ranges()

        curextract = 0
        data = {}
        for c in self.cycles:
            hfile = os.path.join(self.dirname, "{}.cycle_{}/domain_000000.hdf5".format(self.basename, c))
            with h5py.File(hfile, "r") as bpf:
                # shape
                w = bpf["coordsets/coords/dims/i"][0] - 1
                h = bpf["coordsets/coords/dims/j"][0] - 1
    
                for v in self.variables:
                    # write the compressed data
                    data[v] = bpf.get("fields/{}/values".format(v))[...].reshape((w,h))
                    numpy.savez_compressed("{}/{}".format(self.outcdb, str(curextract).zfill(6)), data=data[v])
    
                    channel = v
                    if channel == 'depth':
                        channel = 'CISDepth'
                    # insert an entry in to the database
                    id = cdb.add_entry({'cycle':                c, 
                                        'CISImage':             'cycle_{}'.format(c.zfill(6)), 
                                        'CISVersion':           '1.0', 
                                        'CISOrigin':            'UL', 
                                        'CISImageWidth':        w,
                                        'CISImageHeight':       h,
                                        'CISLayer':             'layer0',
                                        'CISLayerOffsetX':      0, 
                                        'CISLayerOffsetY':      0, 
                                        'CISLayerWidth':        w, 
                                        'CISLayerHeight':       h, 
                                        'CISChannel':           channel, 
                                        'CISChannelVar':        v, 
                                        'CISChannelVarType':    'float', 
                                        'CISChannelVarMin':     self.minmax[v][0], 
                                        'CISChannelVarMax':     self.minmax[v][1],
                                        'FILE':                 '{}.npz'.format(str(curextract).zfill(6))
                                       })
                    curextract = curextract + 1
    
        cdb.finalize()

    #
    # run through the data, computing the min/max of all variables
    #
    def __get_variable_ranges(self):

        for c in self.cycles:
            hfile = os.path.join(self.dirname, "{}.cycle_{}/domain_000000.hdf5".format(self.basename, c))
            with h5py.File(hfile, "r") as bpf:
                w = bpf["coordsets/coords/dims/i"][0] - 1
                h = bpf["coordsets/coords/dims/j"][0] - 1

                # get the variable names
                fields = bpf["fields"]
                for val in fields.keys():
                    if not val == "ascent_ghosts":
                        if not val in self.variables:
                            self.variables.append(val)

                # get the min/max variables
                for v in self.variables:
                    data = bpf.get("fields/{}/values".format(v))[...].reshape((w,h))
                    if v in self.minmax:
                        vmin = numpy.nanmin(data)
                        vmax = numpy.nanmax(data)
                        if vmin < self.minmax[v][0]:
                            self.minmax[v][0] = vmin
                        if vmax > self.minmax[v][1]:
                            self.minmax[v][1] = vmax
                    else:
                        self.minmax[v] = [numpy.nanmin(data), numpy.nanmax(data)]

    def __get_cycles(self):
        cycles = []
        hpaths = glob.glob("{}/*.cycle_[0-9]*.root".format(self.indir))

        self.dirname   = os.path.dirname(hpaths[0])
        self.basename  = os.path.basename(hpaths[0]).split('.')[0]

        for p in hpaths:
            split_path = p.split('_')
            cycle = split_path[-1].split('.')
            self.cycles.append(cycle[0])

        self.cycles.sort()



r"""
This module provides utility functions that can be used to expose a
`VTKPythonAlgorithmBase` derived Pythonic vtkAlgorithm subclass in ParaView.

Introduction
============

VTK provides a convenient Python-friendly mechanism to add new algorithms, such as data sources,
filters, readers, and writers via `vtkPythonAlgorithm`. In this approach, developers can
write subclasses of `VTKPythonAlgorithmBase` and then implement the
`vtkAlgorithm` API in Python.

.. code-block:: python

    from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
    class ContourShrink(VTKPythonAlgorithmBase):
      def __init__(self):
          VTKPythonAlgorithmBase.__init__(self)

      def RequestData(self, request, inInfo, outInfo):
          inp = vtk.vtkDataSet.GetData(inInfo[0])
          opt = vtk.vtkPolyData.GetData(outInfo)

          cf = vtk.vtkContourFilter()
          cf.SetInputData(inp)
          cf.SetValue(0, 200)

          sf = vtk.vtkShrinkPolyData()
          sf.SetInputConnection(cf.GetOutputPort())
          sf.Update()

          opt.ShallowCopy(sf.GetOutput())
          return 1


Such `VTKPythonAlgorithmBase`-based algorithms act as any other `vtkAlgorithm`
and can be used directly in VTK pipelines. To use them in ParaView, however, one
has to make ParaView aware of the capabilities of the algorithm. For
`vtkAlgorithm` subclasses in C++, this is done by writing an XML description of
the algorithm. For Python-based algorithms, this module provides us a mechanism
to build the ParaView-specific XML via decorators.

.. code-block:: python

    from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
    from paraview.util.vtkAlgorithm import smproxy, smproperty, smdomain

    @smproxy.filter(label="Contour Shrink")
    @smproperty.input(name="Input")
    @smdomain.datatype(dataTypes=["vtkDataSet"], composite_data_supported=False)
    class ContourShrink(VTKPythonAlgorithmBase):
      def __init__(self):
          VTKPythonAlgorithmBase.__init__(self)

      def RequestData(self, request, inInfo, outInfo):
          ...
          return 1


.. note::

    ParaView's XML description provides a wide gamut of features.
    The decorators available currently only expose a small subset of them. The
    plan is to have complete feature parity in the future. If you encounter a
    missing capability, please raise an issue or better yet, a merge-request to address
    the same.


Decorator Basics
================

There are four main classes currently available: `smproxy`,
`smproperty`, `smdomain`, and `smhint`. Each provides a set of decorators to
declare a proxy, property, domain or a hint, respectively.

All available decorators exhibit the following characteristics:

#. With the exception of decorators defined in `smproxy`, all decorators are
   intended for decorating classes or their methods. `smproxy` decorators are
   only intended for classes.
#. Decorators take keyword arguments (positional arguments are discouraged).
   While there may be exceptions, the general pattern is that a decorator
   corresponds to an XML node in the ParaView XML configuration. Any attributes
   on the XML node itself can be specified as keyword arguments to the
   decorator. The decorator may explicitly define parameters that affect the
   generated node or to explicit limit the arguments provided.
#. Decorators can be chained. The order of chaining is significant and follows
   the direction of nesting in the XML e.g. to add a domain to a property, one
   would nest decorators as follows:

    .. code-block:: python

        @smproperty.intvector(name="Radius", default_values=[0])
        @smdomain.intrange(min=0, max=10)
        @smhint.xml("<SomeHint/>")
        def SetRadius(self, radius):
            pass

   Note, since domains can't have nested hints, the hint specified after the
   `smdomain.intrange` invocation is intended for the `smproperty` and not the
   `smdomain`.

   When chaining `smproperty` decorators under a `smproxy`, the order in which
   the properties are shown in ParaView UI is reverse of the order in the
   Python code e.g. the in following snippets the UI will show "PolyDataInput"
   above "TableInput".

.. code-block:: python

    @smproxy.filter(...)
    @smproperty.input(name="TableInput", port_index=1, ...)
    @smdomain.datatype(dataTypes=["vtkTable"])
    @smproperty.input(name="PolyDataInput", port_index=0, ...)
    @smdomain.datatype(dataTypes=["vtkPolyData"])
    class MyFilter(VTKPythonAlgorithmBase):
       ...


`smproxy` Decorators
====================

To make `VTKPythonAlgorithmBase` subclass available in ParaView as a data-source,
filter, reader, or writer, one must use of the decorators in the `smproxy`
class.

Common decorator parameters
---------------------------

The following keyword parameters are supported by all `smproxy` decorators:

#. `name`: if present, provides the name to use of the Proxy. This is the name
   that ParaView uses to refer to your class. If none specified, the name is
   deduced from the class name.
#. `label`: if present, provides the label to use in the UI. If missing, the
   label is deduced from the `name`.
#. `class`: if present, is the fully qualified class name ParaView should use to
   instantiate this object. Except for exceptional situation, it's best to let
   ParaView deduce this argument.
#. `group`: if present, is the proxy-group under which the definition must be
   placed. This is often not needed since the decorators set a good default
   based on the role of the proxy.

As mentioned before, any keyword arguments passed any decorator that is not
processed by the decorator simply gets added to the generated XML node as
attribute.

`smproxy.source`
----------------

This decorator can be used to declare data sources. Data sources don't have any
inputs, but provide 1 or more outputs. While technically a reader is a source,
`smproxy.reader` is preferred way of adding readers.

`smproxy.filter`
----------------

This decorator can be used to declare filters that have one or more inputs and
one or more outputs. The inputs themselves must be declared using
`smproperty.input` decorators.

`smproxy.reader`
----------------

This decorator is used to declare reader. This is same as `smproxy.source`,
except it adds the appropriate hints to make ParaView aware that this is a
reader.

It takes the following keyword parameters, in addition to the standard ones:

#. `file_description`: (required) a user friendly text to use in the **File
   Open** dialog.
#. `extensions`:  a string or list of strings which define the supported
   extensions (without the leading `.`).
#. `filename_patterns`: wild card patten.

Either `extensions` or `filename_patterns` must be provided.

`smproxy.writer`
----------------

This decorator is used to declare a writer. Similar to `smproxy.reader`, one
provides `file_description` and `extensions` keyword parameters to indicate
which file extensions the writer supports.


`smproperty` Decorators
=======================

Decorators in `smproperty` are used to add properties to the proxies declared
using `smproxy` decorators. These can be added either to class methods or to
class objects. Multiple `smproperty` decorators may be chained when adding to
class objects.

Common decorator parameters
---------------------------

The following keyword parameters are supported by all `smproperty` decorators:

#. `name`: if present, provides the name to use of the Property. If missing, the
   name will be deduced from the object being decorated which will rarely be
   correct when decorating class objects. Hence, this parameter is optional
   when decorating class methods, but required when decorating class objects.
#. `label`: if present, provides the label to use in the UI. If missing, the
   label is deduced from the `name`.
#. `command`: if present, is the method name. If missing, ParaView can deduce it
   based on the method being decorated. Hence, as before, optional when
   decorating class methods, but generally required when decorating class
   objects (with the exception of `smproperty.input`).

`smproperty.xml`
----------------

This is a catch-all decorator. One can use this to add arbitrary XML to the
proxy definition.

.. code-block:: python

    @smproperty.xml(xmlstr="<IntVectorProperty name=..> ....  </IntVectorProperty>")
    def MyMethod(self, ...):
        ...


`smproperty.intvector`, `smproperty.doublevector`, `smproperty.idtypevector`, `smproperty.stringvector`
-------------------------------------------------------------------------------------------------------

These are used to declare various types of vector properties. The arguments are
simply the XML attributes.

`smproperty.proxy`
------------------

This decorator can be used for adding a `ProxyProperty`, i.e. a method that
takes another VTK object as argument.

`smproperty.input`
------------------

This decorator can be used to add inputs to a `vtkAlgorithm`. Since
`vtkAlgorithm` provides the appropriate API to set an input, often one doesn't
have custom methods in Python in the `VTKPythonAlgorithmBase` subclass. Hence,
this decorator is often simply added to the class definition rather than a
particular method.

.. code-block:: python

    @smproxy.filter(...)
    @smproperty.input(name="TableInput", port_index=0, multiple_input=True)
    @smdomain.datatype(dataTypes=["vtkTable"], composite_data_supported=False)
    class MyFilter(VTKPythonAlgorithmBase):
       ...

`smproperty.dataarrayselection`
-------------------------------

This a convenience decorator that can be used to decorate a method that returns
a `vtkDataArraySelection` object often used to array selection in data sources
and readers. The decorator adds appropriate ParaView ServerManager XML
components to expose the array selection to the user.


`smdomain` Decorators
=======================

These decorators are used to add domains to `smproperty`. Domains often guide
the UI that ParaView generated for the specific property. The current set of
available domains, includes `smdomain.intrange`, `smdomain.doublerange`,
`smdomain.filelist`, and `smdomain.datatype`. More will be added as needed.
Additionally, one can use the `smdomain.xml` catch-all decorator to add XML
definition for the domain.


`smhint` Decorators
====================

These add elements under `<Hints/>` for `smproxy` and/or `smproperty` elements.
Currently `smhint.filechooser` and the generic `smhint.xml` are provided with
more to be added as needed.


Examples
========

**PythonAlgorithmExamples.py** provides a working example that demonstrates how
several of the decorators described here may be used.

.. literalinclude:: ../PythonAlgorithmExamples.py
   :caption:

"""
from __future__ import absolute_import, print_function

from ..detail.pythonalgorithm import \
    smproxy, \
    smproperty, \
    smdomain, \
    smhint

# import VTKPythonAlgorithmBase from vtkmodules, just to make it easier for
# importers of this module.
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase



import xml.etree.ElementTree as ET
from urllib.request import urlopen
import os

class colormap:
    """Colormap Class
    A layer is a collection of values that comprise of the colormap. 
    A colormap is usually a list of object, each object in the form (x, r, g, b) where x is a value 
    0 and 1 and r,g,b normalized values between zero and one that correspond to the red, green and blue 
    color channels. 
    Are we considering alpha channels in colormaps?
    What is the dimension of a colormap?
    """

    def __init__(self, pathToFile):
        """Colormap constructor"""
        self.pathToFile = pathToFile
        self.name = os.path.splitext(os.path.basename(pathToFile))[0]
        self.points = []
        self.typeXML = True
        self.edited = False

        #todo: check if file exists
        urlCheck = pathToFile[0:4]
        if (urlCheck == 'http'):
            self.typeXML = False
            pathToFile = urlopen(pathToFile)
        tree = ET.parse(pathToFile)

        self.root = tree.getroot()
        self.__indent(self.root)

#        for cmap in self.root.findall('ColorMap'):
#            self.name = cmap.get('name')

        for point in self.root.iter('Point'):
            value = point.get('x')
            alpha = point.get('o')
            red   = point.get('r')
            green = point.get('g')
            blue  = point.get('b')
            self.points.append((float(value), float(alpha),
                                float(red), float(green), float(blue)))

    def __indent(self, elem, level=0):
        i = "\n" + level*"  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self.__indent(elem, level+1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i


    def get_points(self):
        """Return a list of tuples that define the colormap points
        """
        return self.points

    def add_point(self, point):
        """ Add a point, a tuple of (x,o,r,g,b) to points. """
        self.edited = True
        self.points.append(point)
        self.points.sort()
        #to do, if x value of point already exists, replace or return error?
        return

    def remove_point(self, point):
        """ Remove a point, a tuple of (x,o,r,g,b) from points. """
        self.edited = True
        if point in points:
            self.points.remove(point)
        # if point not in points, indicate that with an error?
        return





r"""simple is a module for using paraview server manager in Python. It
provides a simple convenience layer to functionality provided by the
C++ classes wrapped to Python as well as the servermanager module.

A simple example::

.. code:: python

    from paraview.simple import *

    # Create a new sphere proxy on the active connection and register it
    # in the sources group.
    sphere = Sphere(ThetaResolution=16, PhiResolution=32)

    # Apply a shrink filter
    shrink = Shrink(sphere)

    # Turn the visibility of the shrink object on.
    Show(shrink)

    # Render the scene
    Render()

Scripts generated by ParaView's Python Tracing, Python state saving, and
Catalyst saving use functions in this module.
"""

import paraview
from paraview import servermanager

from paraview.servermanager import (
    OutputPort,
)

from paraview.simple.version import (
    GetOpenGLInformation,
    GetParaViewSourceVersion,
    GetParaViewVersion,
)
from paraview.simple.session import (
    Connect,
    Disconnect,
    ResetSession,
    ReverseConnect,
    SetActiveConnection,
    GetActiveView,
    GetActiveSource,
    GetActiveCamera,
    SetActiveSource,
    SetActiveView,
    _add_functions,
    _DisableFirstRenderCameraReset,
)
from paraview.simple.proxy import (
    RenameProxy,
    RenameSource,
    RenameView,
    FindSource,
    FindView,
    GetActiveViewOrCreate,
    FindViewOrCreate,
    LocateView,
    GetSources,
    GetSelections,
    GetRepresentations,
    Delete,
    ResetProperty,
    UpdatePipeline,
)
from paraview.simple.plugin import (
    LoadXML,
    LoadPlugins,
    LoadPlugin,
    LoadDistributedPlugin,
    LoadCustomFilters,
)
from paraview.simple.io import (
    LoadState,
    SaveState,
    OpenDataFile,
    ReloadFiles,
    ExtendFileSeries,
    ReplaceReaderFileName,
    CreateWriter,
    SaveData,
    SaveScreenshot,
    SetNumberOfCallbackThreads,
    GetNumberOfCallbackThreads,
    SetNumberOfSMPThreads,
    GetNumberOfSMPThreads,
    WaitForScreenshot,
    SaveAnimation,
    WriteAnimationGeometry,
    FetchData,
)
from paraview.simple.view import (
    CreateView,
    CreateRenderView,
    CreateXYPlotView,
    CreateXYPointPlotView,
    CreateBarChartView,
    CreateComparativeRenderView,
    CreateComparativeXYPlotView,
    CreateComparativeBarChartView,
    CreateParallelCoordinatesChartView,
    Create2DRenderView,
    GetRenderView,
    GetRenderView,
    GetRenderViews,
    GetViews,
    SetViewProperties,
    ExportView,
    ImportView,
)
from paraview.simple.rendering import (
    Render,
    RenderAllViews,
    Interact,
    GetRepresentation,
    GetDisplayProperties,  # to deprectate
    Show,
    ShowAll,
    Hide,
    HideAll,
    SetRepresentationProperties,
    SetDisplayProperties,  # to deprecate
    ColorBy,
    ColorBlocksBy,
)

from paraview.simple.camera import (
    ResetCamera,
    ResetCameraToDirection,
)
from paraview.simple.layout import (
    RenameLayout,
    CreateLayout,
    RemoveLayout,
    GetLayouts,
    GetLayout,
    GetLayoutByName,
    GetViewsInLayout,
    AssignViewToLayout,
    RemoveViewsAndLayouts,
    EqualizeViewsHorizontally,
    EqualizeViewsVertically,
    EqualizeViewsBoth,
)
from paraview.simple.color import (
    LoadPalette,
    GenerateRGBPoints,
    GetColorTransferFunction,
    GetBlockColorTransferFunction,
    GetOpacityTransferFunction,
    GetTransferFunction2D,
    ImportPresets,
    ExportTransferFunction,
    CreateLookupTable,
    CreatePiecewiseFunction,
    AssignFieldToColorPreset,
    ListColorPresetNames,
)

from paraview.simple.link import (
    RemoveLink,
    AddProxyLink,
    AddViewLink,
    AddRenderViewLink,
    AddCameraLink,
    RemoveCameraLink,
    AddSelectionLink,
    RemoveSelectionLink,
)
from paraview.simple.animation import (
    GetTimeKeeper,
    GetAnimationScene,
    AnimateReader,
    GetRepresentationAnimationHelper,
    GetAnimationTrack,
    GetCameraTrack,
    GetTimeTrack,
)
from paraview.simple.light import (
    CreateLight,
    AddLight,
    RemoveLight,
    GetLight,
    GetViewForLight,
)
from paraview.simple.widget import (
    HideUnusedScalarBars,
    HideScalarBarIfNotNeeded,
    UpdateScalarBars,
    UpdateScalarBarsComponentTitle,
    GetScalarBar,
    ShowInteractiveWidgets,
    HideInteractiveWidgets,
)
from paraview.simple.selection import (
    SelectionProxy,
    CreateSelection,
    SelectSurfacePoints,
    SelectSurfaceCells,
    SelectSurfaceBlocks,
    SelectPointsThrough,
    SelectCellsThrough,
    SelectGlobalIDs,
    SelectPedigreeIDs,
    SelectIDs,
    SelectCompositeDataIDs,
    SelectHierarchicalDataIDs,
    SelectThresholds,
    SelectLocation,
    QuerySelect,
    SelectCells,
    SelectPoints,
    ClearSelection,
)
from paraview.simple.texture import (
    GetMaterialLibrary,
    CreateTexture,
    FindTextureOrCreate,
)
from paraview.simple.catalyst import (
    CreateExtractor,
    GetExtractors,
    FindExtractor,
    SaveExtractsUsingCatalystOptions,
    SaveExtracts,
    CreateSteerableParameters,
    UpdateSteerableParameters,
)
from paraview.simple.setting import (
    GetSettingsProxy,
    GetAllSettings,
)

from paraview.simple.deprecated import (
    MakeBlueToRedLT,
    SetProperties,
    GetProperty,
    GetRepresentationProperty,
    GetDisplayProperty,
    GetViewProperties,
    GetViewProperty,
    AssignLookupTable,
    GetLookupTableNames,
    LoadLookupTable,
)

__all__ = [
    "_extend_simple",
    # -- servermanager
    "paraview",  # needed to do `from paraview.simple import *;paraview.simple._DisableFirstRenderCameraReset()`
    "servermanager",  # needed for legacy import usage
    "OutputPort",
    # -- version
    "GetOpenGLInformation",
    "GetParaViewSourceVersion",
    "GetParaViewVersion",
    # -- session
    "Connect",
    "Disconnect",
    "ResetSession",
    "ReverseConnect",
    "SetActiveConnection",
    "_DisableFirstRenderCameraReset",
    # -- active
    "GetActiveView",
    "GetActiveSource",
    "GetActiveCamera",
    "SetActiveSource",
    "SetActiveView",
    # -- proxy
    "RenameProxy",
    "RenameSource",
    "RenameView",
    "FindSource",
    "FindView",
    "GetActiveViewOrCreate",
    "FindViewOrCreate",
    "LocateView",
    "GetSources",
    "GetSelections",
    "GetRepresentations",
    "Delete",
    "ResetProperty",
    "UpdatePipeline",
    # -- plugin
    "LoadXML",
    "LoadPlugins",
    "LoadPlugin",
    "LoadDistributedPlugin",
    "LoadCustomFilters",
    # -- io
    "LoadState",
    "SaveState",
    "OpenDataFile",
    "ReloadFiles",
    "ExtendFileSeries",
    "ReplaceReaderFileName",
    "CreateWriter",
    "SaveData",
    "SaveScreenshot",
    "SetNumberOfCallbackThreads",
    "GetNumberOfCallbackThreads",
    "SetNumberOfSMPThreads",
    "GetNumberOfSMPThreads",
    "WaitForScreenshot",
    "SaveAnimation",
    "WriteAnimationGeometry",
    "FetchData",
    # -- view
    "CreateView",
    "CreateRenderView",
    "CreateXYPlotView",
    "CreateXYPointPlotView",
    "CreateBarChartView",
    "CreateComparativeRenderView",
    "CreateComparativeXYPlotView",
    "CreateComparativeBarChartView",
    "CreateParallelCoordinatesChartView",
    "Create2DRenderView",
    "GetRenderView",
    "GetRenderView",
    "GetRenderViews",
    "GetViews",
    "SetViewProperties",
    "ExportView",
    "ImportView",
    # -- rendering
    "Render",
    "RenderAllViews",
    "Interact",
    "GetRepresentation",
    "GetDisplayProperties",
    "Show",
    "ShowAll",
    "Hide",
    "HideAll",
    "SetRepresentationProperties",
    "SetDisplayProperties",
    "ColorBy",
    "ColorBlocksBy",
    # -- camera
    "ResetCamera",
    "ResetCameraToDirection",
    # -- layout
    "RenameLayout",
    "CreateLayout",
    "RemoveLayout",
    "GetLayouts",
    "GetLayout",
    "GetLayoutByName",
    "GetViewsInLayout",
    "AssignViewToLayout",
    "RemoveViewsAndLayouts",
    "EqualizeViewsHorizontally",
    "EqualizeViewsVertically",
    "EqualizeViewsBoth",
    # -- color
    "LoadPalette",
    "GenerateRGBPoints",
    "GetColorTransferFunction",
    "GetBlockColorTransferFunction",
    "GetOpacityTransferFunction",
    "GetTransferFunction2D",
    "ImportPresets",
    "ExportTransferFunction",
    "CreateLookupTable",
    "CreatePiecewiseFunction",
    "AssignFieldToColorPreset",
    "ListColorPresetNames",
    # -- link
    "RemoveLink",
    "AddProxyLink",
    "AddViewLink",
    "AddRenderViewLink",
    "AddCameraLink",
    "RemoveCameraLink",
    "AddSelectionLink",
    "RemoveSelectionLink",
    # -- animation
    "GetTimeKeeper",
    "GetAnimationScene",
    "AnimateReader",
    "GetRepresentationAnimationHelper",
    "GetAnimationTrack",
    "GetCameraTrack",
    "GetTimeTrack",
    # -- light
    "CreateLight",
    "AddLight",
    "RemoveLight",
    "GetLight",
    "GetViewForLight",
    # -- widget
    "HideUnusedScalarBars",
    "HideScalarBarIfNotNeeded",
    "UpdateScalarBars",
    "UpdateScalarBarsComponentTitle",
    "GetScalarBar",
    "ShowInteractiveWidgets",
    "HideInteractiveWidgets",
    # -- selection
    "SelectionProxy",
    "CreateSelection",
    "SelectSurfacePoints",
    "SelectSurfaceCells",
    "SelectSurfaceBlocks",
    "SelectPointsThrough",
    "SelectCellsThrough",
    "SelectGlobalIDs",
    "SelectPedigreeIDs",
    "SelectIDs",
    "SelectCompositeDataIDs",
    "SelectHierarchicalDataIDs",
    "SelectThresholds",
    "SelectLocation",
    "QuerySelect",
    "SelectCells",
    "SelectPoints",
    "ClearSelection",
    # -- texture
    "GetMaterialLibrary",
    "CreateTexture",
    "FindTextureOrCreate",
    # -- catalyst
    "CreateExtractor",
    "GetExtractors",
    "FindExtractor",
    "SaveExtractsUsingCatalystOptions",
    "SaveExtracts",
    "CreateSteerableParameters",
    "UpdateSteerableParameters",
    # -- setting
    "GetSettingsProxy",
    "GetAllSettings",
    # -- compatibility
    # -- deprecated
    "MakeBlueToRedLT",
    "SetProperties",
    "GetProperty",
    "GetRepresentationProperty",
    "GetDisplayProperty",
    "GetViewProperties",
    "GetViewProperty",
    "AssignLookupTable",
    "GetLookupTableNames",
    "LoadLookupTable",
]

# ==============================================================================
# Namespace extension handling
# ==============================================================================


def _extend_simple(*namespaces):
    local_ns = globals()
    for name in _add_functions(local_ns):
        if name not in __all__:
            __all__.append(name)

    for ns in namespaces:
        if ns is not local_ns:
            for name in _add_functions(ns):
                if name not in __all__:
                    __all__.append(name)


# ==============================================================================
# Start the session and initialize the ServerManager
# ==============================================================================

if not paraview.options.satellite:
    if not servermanager.ActiveConnection:
        Connect()
    else:
        # needed for catalyst pvbatch mpi
        _extend_simple()



"""
Test paraview python backward compatibility helper for deprecated proxies.
We use GhostCellsGenerator that is deprecated in 5.12 and renamed as GhostCells.
"""
import paraview
paraview.compatibility.major = 5
paraview.compatibility.minor = 12

from paraview.simple import *

def main():
    wavelet1 = Wavelet()

    # We test if the ghost cells generator can be instantiated
    generator = GhostCellsGenerator(Input=wavelet1)

    # Verify that the concrete proxy is the new one
    if type(generator).__name__ != "GhostCells":
        print("error, proxy not substituted: is of type ", type(generator).__name__)

if __name__ == "__main__":
    main()



r"""web_helper is a module that provides access to functions that helps to build
new protocols and process ParaView data structure into web friendly ones.
"""

import sys
import os
import traceback

# import paraview modules.
import paraview

from paraview import simple, servermanager
from paraview.servermanager import ProxyProperty
from paraview.modules.vtkRemotingViews import vtkSMPVRepresentationProxy

# =============================================================================
# Pipeline management
# =============================================================================


class Pipeline:
    """
    Define a data structure that represent a pipeline as a tree.
    This provide also methods to get the data structure for the web environment
    """

    # --------------------------------------------------------------------------

    def __init__(self, name):
        self.root_node = {"name": name, "icon": "server", "children": []}
        self.parent_ids = {"0": "0"}
        self.children_ids = {"0": []}

    # --------------------------------------------------------------------------

    def clear(self):
        """
        Clear the pipeline tree.
        """
        self.root_node["children"] = []
        self.parent_ids = {"0": "0"}
        self.children_ids = {"0": []}

    # --------------------------------------------------------------------------

    def addNode(self, parent_id, node_id):
        """
        Add node into the pipeline tree.
        """
        pid = str(parent_id)
        nid = str(node_id)

        # Add child
        if pid in self.children_ids:
            self.children_ids[pid].append(nid)
        else:
            self.children_ids[pid] = [nid]

        # Add parent
        self.parent_ids[nid] = pid

    # --------------------------------------------------------------------------

    def removeNode(self, id):
        """
        Remove a node from the pipeline tree.
        """
        nid = str(id)
        pid = self.parent_ids[nid]
        if pid:
            del self.parent_ids[nid]
            self.children_ids[pid].remove(nid)

    # --------------------------------------------------------------------------

    def isEmpty(self):
        return len(self.parent_ids) == 1

    # --------------------------------------------------------------------------

    def getRootNode(self, view=None):
        """
        Create a tree structure of the pipeline with the current proxy state.
        """
        self.root_node["children"] = []
        self.__fill_children(self.root_node, self.children_ids["0"], view)
        return self.root_node

    # --------------------------------------------------------------------------

    def __fill_children(self, nodeToFill, childrenIds, view=None):
        for id in childrenIds:
            node = getProxyAsPipelineNode(id, view)
            nid = str(node["proxy_id"])

            if "children" in nodeToFill:
                nodeToFill["children"].append(node)
            else:
                nodeToFill["children"] = [node]

            if nid in self.children_ids:
                self.__fill_children(node, self.children_ids[nid])


# =============================================================================
# Proxy management
# =============================================================================


def idToProxy(id):
    """
    Return the proxy that match the given proxy ID.
    """
    remoteObject = simple.servermanager.ActiveConnection.Session.GetRemoteObject(
        int(id)
    )
    if remoteObject:
        return simple.servermanager._getPyProxy(remoteObject)
    return None


# --------------------------------------------------------------------------


def getParentProxyId(proxy):
    """
    Return '0' if the given proxy has no Input otherwise will return
    the parent proxy id as a String.
    """
    if proxy and proxy.GetProperty("Input"):
        parentProxy = proxy.GetProperty("Input").GetProxy(0)
        if parentProxy:
            return parentProxy.GetGlobalIDAsString()
    return "0"


# --------------------------------------------------------------------------


def getProxyAsPipelineNode(id, view=None):
    """
    Create a representation for that proxy so it can be used within a pipeline
    browser.
    """
    pxm = servermanager.ProxyManager()
    proxy = idToProxy(id)
    rep = simple.GetDisplayProperties(proxy)
    nbActiveComp = 1

    pointData = []
    searchArray = ("POINTS" == rep.ColorArrayName[0]) and (
        len(rep.ColorArrayName[1]) > 0
    )

    if servermanager.ActiveConnection.GetNumberOfDataPartitions() > 1:
        info = {
            "lutId": "vtkProcessId_1",
            "name": "vtkProcessId",
            "size": 1,
            "range": [
                0,
                servermanager.ActiveConnection.GetNumberOfDataPartitions() - 1,
            ],
        }
        pointData.append(info)

    # FIXME seb
    # dataInfo = rep.GetRepresentedDataInformation()
    # pointData = dataInfo.GetPointDataInformation()
    # cellData = dataInfo.GetCellDataInformation()
    # for idx in pointData.GetNumberOfArrays():
    #     info = pointData.GetArrayInformation(idx)
    #     nbComponents = info.GetNumberOfComponents()
    #     if searchArray and array.Name == rep.ColorArrayName:
    #         nbActiveComp = nbComponents
    #     rangeOn = (nbComponents == 3 if -1 else 0)
    #     info = {                                      \
    #     'lutId': info.GetName() + '_' + str(nbComponents), \
    #     'name': info.GetName,                             \
    #     'size': nbComponents,                            \
    #     'range': info.GetRange(rangeOn) }
    #     pointData.append(info)

    for array in proxy.GetPointDataInformation():
        nbComponents = array.GetNumberOfComponents()
        if searchArray and array.Name == rep.ColorArrayName[1]:
            nbActiveComp = nbComponents
        rangeOn = nbComponents == 1 if 0 else -1
        info = {
            "lutId": array.Name + "_" + str(nbComponents),
            "name": array.Name,
            "size": nbComponents,
            "range": array.GetRange(rangeOn),
        }
        pointData.append(info)

    cellData = []
    searchArray = ("CELLS" == rep.ColorArrayName[0]) and (
        len(rep.ColorArrayName[1]) > 0
    )
    for array in proxy.GetCellDataInformation():
        nbComponents = array.GetNumberOfComponents()
        if searchArray and array.Name == rep.ColorArrayName[1]:
            nbActiveComp = nbComponents
        rangeOn = nbComponents == 1 if 0 else -1
        info = {
            "lutId": array.Name + "_" + str(nbComponents),
            "name": array.Name,
            "size": nbComponents,
            "range": array.GetRange(rangeOn),
        }
        cellData.append(info)

    state = getProxyAsState(proxy.GetGlobalID())
    showScalarbar = (
        1
        if view
        and vtkSMPVRepresentationProxy.IsScalarBarVisible(rep.SMProxy, view.SMProxy)
        else 0
    )

    repName = "Hide"
    if rep.Visibility == 1:
        repName = rep.Representation

    return {
        "proxy_id": proxy.GetGlobalID(),
        "name": pxm.GetProxyName("sources", proxy),
        "bounds": proxy.GetDataInformation().GetBounds(),
        "pointData": pointData,
        "cellData": cellData,
        "activeData": str(rep.ColorArrayName[0]) + ":" + str(rep.ColorArrayName[1]),
        "diffuseColor": str(rep.DiffuseColor),
        "showScalarBar": showScalarbar,
        "representation": repName,
        "state": state,
        "children": [],
    }


# --------------------------------------------------------------------------


def getProxyAsState(id):
    """
    Return a json representation of the given proxy state.

    Example of the state of the Clip filter:

    .. code-block:: json

        {
            "proxy_id": 234,
            "ClipType": {
                "proxy_id": 235,
                "Normal": [0, 0, 1],
                "Origin": [0, 0, 0],
                "InsideOut": 0
            }
        }

    """
    proxy_id = int(id)
    proxy = idToProxy(proxy_id)
    state = {
        "proxy_id": proxy_id,
        "type": "proxy",
        "domains": getProxyDomains(proxy_id),
    }
    properties = {}
    allowedTypes = [int, float, list, str]
    if proxy:
        for property in proxy.ListProperties():
            propertyName = proxy.GetProperty(property).Name
            if propertyName in ["Refresh", "Input"] or propertyName.__contains__(
                "Info"
            ):
                continue

            data = proxy.GetProperty(property).GetData()
            if type(data) in allowedTypes:
                properties[propertyName] = data
                continue

            # Not a simple property
            # Need more investigation
            prop = proxy.GetProperty(property)
            pythonProp = servermanager._wrap_property(proxy, prop)
            proxyList = []
            try:
                proxyList = pythonProp.Available
            except:
                pass
            if len(proxyList) and prop.GetNumberOfProxies() == 1:
                listdomain = prop.FindDomain("vtkSMProxyListDomain")
                if listdomain:
                    proxyPropertyValue = prop.GetProxy(0)
                    for i in range(listdomain.GetNumberOfProxies()):
                        if listdomain.GetProxy(i) == proxyPropertyValue:
                            properties[propertyName] = proxyList[i]
                            # Add selected proxy in list of prop to edit
                            properties[propertyName + "_internal"] = getProxyAsState(
                                listdomain.GetProxy(i).GetGlobalID()
                            )

            elif type(prop) == ProxyProperty:
                try:
                    subProxyId = proxy.GetProperty(property).GetData().GetGlobalID()
                    properties[propertyName] = getProxyAsState(subProxyId)
                except:
                    print("Error on", property, propertyName)
                    print("Skip property: ", str(type(data)))
                    print(data)
    state["properties"] = properties
    return state


# --------------------------------------------------------------------------


def updateProxyProperties(proxy, properties):
    """
    Loop over the properties object and update the mapping properties
    to the given proxy.
    """
    try:
        allowedProperties = proxy.ListProperties()
        for key in properties:
            validKey = servermanager._make_name_valid(key)
            if validKey in allowedProperties:
                value = removeUnicode(properties[key])
                property = servermanager._wrap_property(
                    proxy, proxy.GetProperty(validKey)
                )
                if (
                    property.FindDomain("vtkSMProxyListDomain")
                    and len(value) == 1
                    and type(value[0]) == str
                ):
                    try:
                        idx = property.GetAvailable().index(value[0])
                        proxyToSet = servermanager._getPyProxy(
                            property.FindDomain("vtkSMProxyListDomain").GetProxy(idx)
                        )
                        property.SetData(proxyToSet)
                    except:
                        traceback.print_stack()
                        pass
                elif value == "vtkProcessId":
                    property.SetElement(0, value)
                else:
                    property.SetData(value)
    except:
        traceback.print_stack()


# --------------------------------------------------------------------------


def removeUnicode(value):
    return value


# =============================================================================
# XML and Proxy Definition for GUI generation
# =============================================================================


def getProxyDomains(id):
    """
    Return a json based structured based on the proxy XML.
    """
    jsonDefinition = {}
    proxy = idToProxy(id)
    xmlElement = servermanager.ActiveConnection.Session.GetProxyDefinitionManager().GetCollapsedProxyDefinition(
        proxy.GetXMLGroup(), proxy.GetXMLName(), None
    )
    nbChildren = xmlElement.GetNumberOfNestedElements()
    for i in range(nbChildren):
        xmlChild = xmlElement.GetNestedElement(i)
        name = xmlChild.GetName()
        if name.__contains__("Property"):
            propName = xmlChild.GetAttribute("name")
            jsonDefinition[propName] = extractProperty(proxy, xmlChild)
            jsonDefinition[propName]["order"] = i

    # Look for proxy properties and their domain
    orderIndex = nbChildren
    for property in proxy.ListProperties():
        if property == "Input":
            continue
        if type(proxy.GetProperty(property)) == ProxyProperty:
            try:
                subProxyId = proxy.GetProperty(property).GetData().GetGlobalID()
                subDomain = getProxyDomains(subProxyId)
                for key in subDomain:
                    jsonDefinition[key] = subDomain[key]
                    jsonDefinition[key]["order"] = orderIndex
                    orderIndex = orderIndex + 1
            except:
                print("(Def) Error on", property, ", skipping it...")
                # print ("(Def) Skip property: ", str(type(data)))

    return jsonDefinition


def extractProperty(proxy, xmlPropertyElement):
    propInfo = {}
    propInfo["name"] = xmlPropertyElement.GetAttribute("name")
    propInfo["label"] = xmlPropertyElement.GetAttribute("label")
    if xmlPropertyElement.GetAttribute("number_of_elements") != None:
        propInfo["size"] = xmlPropertyElement.GetAttribute("number_of_elements")
    propInfo["type"] = xmlPropertyElement.GetName()[:-14]
    propInfo["panel_visibility"] = xmlPropertyElement.GetAttribute("panel_visibility")
    propInfo["number_of_elements"] = xmlPropertyElement.GetAttribute(
        "number_of_elements"
    )
    propInfo["domains"] = []
    if xmlPropertyElement.GetAttribute("default_values") != None:
        propInfo["default_values"] = xmlPropertyElement.GetAttribute("default_values")
    nbChildren = xmlPropertyElement.GetNumberOfNestedElements()
    for i in range(nbChildren):
        xmlChild = xmlPropertyElement.GetNestedElement(i)
        name = xmlChild.GetName()
        if name.__contains__("Domain"):
            propInfo["domains"].append(extractDomain(proxy, propInfo["name"], xmlChild))
    return propInfo


def extractDomain(proxy, propertyName, xmlDomainElement):
    domainObj = {}
    name = xmlDomainElement.GetName()
    domainObj["type"] = name[:-6]

    # Handle Range
    if name.__contains__("RangeDomain"):
        if xmlDomainElement.GetAttribute("min") != None:
            domainObj["min"] = xmlDomainElement.GetAttribute("min")
        if xmlDomainElement.GetAttribute("max") != None:
            domainObj["max"] = xmlDomainElement.GetAttribute("max")

    # Handle Enum
    if name.__contains__("EnumerationDomain"):
        domainObj["enum"] = []
        nbChildren = xmlDomainElement.GetNumberOfNestedElements()
        for i in range(nbChildren):
            xmlChild = xmlDomainElement.GetNestedElement(i)
            if xmlChild.GetName() == "Entry":
                domainObj["enum"].append(
                    {
                        "text": xmlChild.GetAttribute("text"),
                        "value": xmlChild.GetAttribute("value"),
                    }
                )

    # Handle ArrayListDomain
    if name.__contains__("ArrayListDomain"):
        dataType = xmlDomainElement.GetAttribute("attribute_type")
        if dataType == "Scalars":
            domainObj["nb_components"] = 1
        elif dataType == "Vectors":
            domainObj["nb_components"] = 3
        else:
            domainObj["nb_components"] = -1

    # Handle ProxyListDomain
    if name.__contains__("ProxyListDomain"):
        domainObj["list"] = proxy.GetProperty(propertyName).Available

    # Handle Bounds
    if name.__contains__("BoundsDomain"):
        for attrName in ["default_mode", "mode", "scale_factor"]:
            try:
                attrValue = xmlDomainElement.GetAttribute(attrName)
                if attrValue:
                    domainObj[attrName] = attrValue
            except:
                pass

    return domainObj


# =============================================================================
# File Management
# =============================================================================


def listFiles(pathToList):
    """
    Create a tree structure of the given directory that will be understood by
    the pipelineBrowser widget.
    The provided path should not have a trailing '/'.

    .. code-block:: json

        {
            "children": [
                { "name": "fileName.vtk", "path": "/full_path/to_file/fileName.vtk" },
                { "name": "directoryName", "path": "/full_path/to_file/directoryName", "children": [] }
            ]
        }

    """
    global fileList
    if pathToList[-1] == "/":
        pathToList = pathToList[:-1]
    nodeTree = {}
    nodeTree[pathToList] = {"children": []}
    for path, directories, files in os.walk(pathToList):
        parent = nodeTree[path]
        for directory in directories:
            child = {
                "name": directory,
                "path": (path + "/" + directory),
                "children": [],
            }
            nodeTree[path + "/" + directory] = child
            parent["children"].append(child)
        for filename in files:
            child = {"name": filename, "path": (path + "/" + filename)}
            nodeTree[path + "/" + filename] = child
            parent["children"].append(child)
    fileList = nodeTree[pathToList]["children"]
    return fileList


# =============================================================================
# Apply domains
# =============================================================================


def apply_domains(parentProxy, proxy_id):
    """
    Handle bounds domain
    """
    proxy = idToProxy(proxy_id)

    # Call recursively on each sub-proxy if any
    for property_name in proxy.ListProperties():
        prop = proxy.GetProperty(property_name)
        if prop.IsA("vtkSMProxyProperty"):
            try:
                if len(prop.Available) and prop.GetNumberOfProxies() == 1:
                    listdomain = prop.FindDomain("vtkSMProxyListDomain")
                    if listdomain:
                        for i in range(listdomain.GetNumberOfProxies()):
                            internal_proxy = listdomain.GetProxy(i)
                            apply_domains(
                                parentProxy, internal_proxy.GetGlobalIDAsString()
                            )
            except:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                print("Unexpected error:", exc_type, " line: ", exc_tb.tb_lineno)

    # Reset all properties to leverage domain capabilities
    for prop_name in proxy.ListProperties():
        try:
            prop = proxy.GetProperty(prop_name)
            iter = prop.NewDomainIterator()
            iter.Begin()
            while not iter.IsAtEnd():
                domain = iter.GetDomain()
                iter.Next()

                try:
                    if domain.IsA("vtkSMBoundsDomain"):
                        domain.SetDomainValues(
                            parentProxy.GetDataInformation().GetBounds()
                        )
                except AttributeError as attrErr:
                    print("Caught exception setting domain values in apply_domains:")
                    print(attrErr)

            prop.ResetToDefault()

            # Need to UnRegister to handle the ref count from the NewDomainIterator
            iter.UnRegister(None)
        except:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print("Unexpected error:", exc_type, " line: ", exc_tb.tb_lineno)

    proxy.UpdateVTKObjects()



from vtkmodules.vtkCommonMisc import vtkResourceFileLocator
import os.path, sys, platform


def find_webapp(appname):
    """Returns the path to the  is web app with given name is found in the package."""

    if platform.system() == "Darwin":
        root = "Resources/"
    else:
        from paraview.servermanager import vtkSMProxyManager
        pv_version = "%d.%d" % (vtkSMProxyManager.GetVersionMajor(), vtkSMProxyManager.GetVersionMinor())
        root = "share/paraview-%s/" % pv_version

    filename = root + "web/%s/server/pvw-%s.py" % (appname, appname)
    locator = vtkResourceFileLocator()
    result = locator.Locate(os.path.dirname(os.path.abspath(__file__)), filename)
    if result:
        return os.path.join(result, filename)
    return None


def find_webappcontents(appname):
    path = find_webapp(appname)
    if not path:
        return None

    root = os.path.dirname(os.path.dirname(path))
    return os.path.join(root, "www")


def load_webapp(appname):
    path = find_webapp(appname)
    if not path:
        return None

    # the web-apps are not proper packages right now, which makes it a little
    # ugly to import them.
    sys.path.insert(0, os.path.dirname(path))
    packagename = os.path.splitext(os.path.basename(path))[0]
    return __import__(packagename)


def get_commandline_args(appname):
    clargs = sys.argv[1:]
    if not "-c" in clargs and not "--content" in clargs:
        clargs.append("-c")
        clargs.append(find_webappcontents(appname))
    return clargs


def start_server(appname, description, module, protocol):
    import argparse

    # Create argument parser
    parser = argparse.ArgumentParser(description=description)

    # Add arguments
    module.server.add_arguments(parser)
    protocol.add_arguments(parser)
    args = parser.parse_args(get_commandline_args(appname))
    protocol.configure(args)

    # Start server
    module.server.start_webserver(options=args, protocol=protocol)



r"""
This script takes in any PVSM file and generates a Catalyst Python state file.
"""

import argparse


def _get_options():
    parser = argparse.ArgumentParser( \
        description="Generate Catalyst Python State")
    parser.add_argument("--output", type=str,
                        help="name for the output file (*.py)", required=True)
    parser.add_argument("--extracts-dir", type=str,
                        help="path to directory where to make the Catalyst script save generated extracts",
                        required=True)
    parser.add_argument("--cinema",
                        help="generate Cinema specification summary for the extracts generated.",
                        action="store_true")
    parser.add_argument("--pvsm", type=str,
                        help="pvsm state file to load", required=True)

    return parser


if __name__ == '__main__':
    parser = _get_options()
    args = parser.parse_args()

    from paraview import simple, catalyst

    simple.LoadState(args.pvsm)

    options = catalyst.Options()
    options.ExtractsOutputDirectory = args.extracts_dir
    if args.cinema:
        options.GenerateCinemaSpecification = 1

    # TODO: add a 'simple' version for this, we don't want users to import
    # anything from 'detail' and hence we don't want that in a demo script.
    import os, os.path

    # remove output if it exists
    if os.path.exists(args.output):
        print("removing existing output file:", args.output)
        os.remove(args.output)

    from paraview.detail.catalyst_export import save_catalyst_state

    save_catalyst_state(args.output, options)

    assert os.path.exists(args.output)



from enum import Enum

class CISPARAMS(Enum):
        image           = "CISImage"
        version         = "CISVersion" 
        imageFlags      = "CISImageFlags"
        imageWidth      = "CISImageWidth"
        imageHeight     = "CISImageHeight"
        layer           = "CISLayer"
        layerOffsetX    = "CISLayerOffsetX"
        layerOffsetY    = "CISLayerOffsetY"
        layerWidth      = "CISLayerWidth"
        layerHeight     = "CISLayerHeight"
        channel         = "CISChannel"
        channelVar      = "CISChannelVar"
        channelVarType  = "CISChannelVarType"
        channelVarMin   = "CISChannelVarMin"
        channelVarMax   = "CISChannelVarMax"
        channelColormap = "CISChannelColormap"
        shadowChannel   = "CISShadow"
        depthChannel    = "CISDepth"

        @staticmethod
        def contains(value):
            if value in [e.value for e in CISPARAMS]: 
                return True
            else:
                return False

class cisview:
    """Composible Image Set View Class
       
       Given a cinema database, this class provides access to 
       CIS data within the database.

       We note that CIS entries are expected to use Cinema's FILE column
       to record any files needed.
    """

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def cis(self, value):
        self._depth = value        

    @property
    def shadow(self):
        return self._shadow

    @shadow.setter
    def cis(self, value):
        self._shadow = value        


    def __init__(self, cdb):
        self.cdb = cdb
        self.parameters = []
        self.CISParams = []

        # state
        self.images = []
        self.layers = [] 
        self.channels = {} 

        # find the CIS parameters that are in the database 
        params = cdb.get_parameter_names()
        for p in params:
            if CISPARAMS.contains(p):
                self.CISParams.append(p)
            else:
                self.parameters.append(p)

        # get layer and channel information
        self.images     = self.__get_image_names()
        self.layers     = self.__get_layer_names()
        self.channels   = self.__get_channel_names()
        self._depth     = self.__query_depth()
        self._shadow    = self.__query_shadow()

    def get_cdb_parameters(self):
        return self.parameters

    def get_cis_parameters(self):
        return self.CISParams

    def __get_image_names(self):
        query = "SELECT DISTINCT CISImage from {}".format(self.cdb.tablename)
        result = self.cdb.execute(query)

        names = []
        for row in result: 
            names.append(str(row[0]))

        return names

    def __get_layer_names(self):
        query = "SELECT DISTINCT CISLayer from {}".format(self.cdb.tablename)
        result = self.cdb.execute(query)

        names = []
        for row in result: 
            names.append(str(row[0]))

        return names

    def __get_channel_names(self):
        results = {} 

        for l in self.layers:
            results[l] = []
            query = "SELECT DISTINCT CISChannel from {} WHERE CISLayer = \'{}\'".format(self.cdb.tablename, l)
            result = self.cdb.execute(query)

            for row in result: 
                results[l].append(str(row[0]))

        return results

    def __query_depth(self):
        query = "SELECT DISTINCT CISChannel from {}".format(self.cdb.tablename)
        result = self.cdb.execute(query)

        names = []
        for row in result: 
            names.append(str(row[0]))

        return "CISDepth" in names 

    def __query_shadow(self):
        query = "SELECT DISTINCT CISChannel from {}".format(self.cdb.tablename)
        result = self.cdb.execute(query)

        names = []
        for row in result: 
            names.append(str(row[0]))

        return "CISShadow" in names 

    def get_image_layers(self, image):
        query = "SELECT DISTINCT CISLayer from {} WHERE CISImage = \'{}\'".format(self.cdb.tablename, image)
        result = self.cdb.execute(query) 

        names = []
        for row in result: 
            names.append(str(row[0]))

        return names

    def get_layer_channels(self, image, layer):
        query = "SELECT DISTINCT CISChannel from {} WHERE CISImage = \'{}\' AND CISLayer = \'{}\'".format(self.cdb.tablename, image, layer)
        result = self.cdb.execute(query) 

        names = []
        for row in result: 
            names.append(str(row[0]))

        return names


    def get_channel_extracts(self, image, layer):
        extracts = []
        for channel in self.channels[layer]:
            ext = self.get_channel_extract(image, layer, channel)
            for e in ext:
                extracts.append(e)

        return extracts

    def get_channel_extract(self, image, layer, channel):
        extracts = []
        params = {
                    "CISImage": image,
                    "CISLayer": layer,
                    "CISChannel": channel 
                 }
        extract = self.cdb.get_extracts(params)

        return extract

    #
    #
    #
    def get_image_parameters(self):
        query = "SELECT CISImageWidth, CISImageHeight from {} LIMIT 1".format(self.cdb.tablename)
        results = self.cdb.execute(query)
        data = {
                    "dims": [results[0][0], results[0][1]]
                }

        return data

    #
    #
    #
    def get_layer_parameters(self, image, layer):
        data = {
                    "dims"  : [0.0, 0.0],
                    "offset": [0.0, 0.0] 
               }

        # both must be present, or they are ignored
        if ("CISLayerWidth" in self.CISParams) and ("CISLayerHeight" in self.CISParams):
            query = "SELECT CISLayerWidth, CISLayerHeight from {} WHERE CISImage = \'{}\' and CISLayer = \'{}\' LIMIT 1".format(self.cdb.tablename, image, layer)
            results = self.cdb.execute(query)
            data["dims"] = [results[0][0], results[0][1]] 

        else:
            params = self.get_image_parameters()
            data["dims"] = params["dims"] 

        # both must be present, or they are ignored
        if ("CISLayerOffsetX" in self.CISParams) and ("CISLayerOffsetY" in self.CISParams):
            query = "SELECT CISLayerOffsetX, CISLayerOffsetY from {} WHERE CISImage = \'{}\' and CISLayer = \'{}\' LIMIT 1".format(self.cdb.tablename, image, layer)
            results = self.cdb.execute(query)
            data["offset"] = [results[0][0], results[0][1]] 

        return data

    #
    # get variable, variable min/max and colormap from a channel if the data exists
    # 
    # Variable: required
    # Min, Max are set to default, if not present (optional)
    # Colormap are set to default, if not present (optional)
    #
    def get_channel_parameters(self, image, layer, channel):
        # set the default colormap value, in preparation for the next logic about the colormap
        data = { 
                    "variable": {
                        "name"  : None, 
                        "type"  : "float",
                        "min"   : -100000.0,
                        "max"   :  100000.0
                    },
                    "colormap": {
                        "type" : "default"
                    }
               }

        if ("CISChannelVar" in self.CISParams):
            query = "SELECT CISChannelVar from {} WHERE CISImage = \'{}\' and CISLayer = \'{}\' and CISChannel = \'{}\'".format(
                        self.cdb.tablename, image, layer, channel)
            results = self.cdb.execute(query)
            data["variable"]["name"] = results[0][0]

        if ("CISChannelVarType" in self.CISParams):
            query = "SELECT CISChannelVarType FROM {} WHERE CISImage = \'{}\' and CISLayer = \'{}\' and CISChannel = \'{}\'".format(
                        self.cdb.tablename, image, layer, channel)
            results = self.cdb.execute(query)

            data["variable"]["type"] = results[0][0]

        if ("CISChannelVarMin" in self.CISParams) and ("CISChannelVarMax" in self.CISParams):
            query = "SELECT CISChannelVarMin, CISChannelVarMax FROM {} WHERE CISImage = \'{}\' and CISLayer = \'{}\' and CISChannel = \'{}\'".format(
                        self.cdb.tablename, image, layer, channel)
            results = self.cdb.execute(query)

            data["variable"]["min"] = results[0][0]
            data["variable"]["max"] = results[0][1] 

        if "CISChannelColormap" in self.CISParams:
            query = "SELECT CISChannelColormap FROM {} WHERE CISImage = \'{}\' and CISLayer = \'{}\' and CISChannel = \'{}\'".format(
                        self.cdb.tablename, image, layer, channel)
            results = self.cdb.execute(query)

            if not results[0][0] is "":
                data["colormap"]["type"] = "url"
                data["colormap"]["url"] = results[0][0]
                    
        return data

    #
    # get the image for a parameter set
    #
    def get_image(self, params):
        query = "SELECT DISTINCT CISImage from {} WHERE ".format(self.cdb.tablename)
        
        first = True
        for key in params: 
            if not first:
                query = query + " AND "
            else:
                first = False
            value = params[key]
            query = query + "{} = \'{}\' ".format(key, value)
        results = self.cdb.execute(query)

        retval = results[0][0]
        if len(results) != 1:
            print("ERROR: incorrect return from image query")

        return retval 




from . import imageview
from . import cisview
from . import renderer
from . import convert


class cis:
    """Composible Image Set Class
    The data structure to hold properties of a Composible Image Set.
    """

    def __init__(self, filename):
        """ The constructor. """
        self.fname          = filename
        self.classname      = "COMPOSABLE_IMAGE_SET"
        self.dims           = [0,0]
        self.flags          = "CONSTANT_CHANNELS"
        self.version        = "1.0"
        self.parameterlist  = []
        self.parametertable = None
        self.variables      = {}
        self.images         = {}
        self.colormaps      = {}

    def debug_print(self):
        """ Debug print statement for CIS properties. """
        print("printing cis")
        print("  fname:     {}".format(self.fname))
        print("  classname: {}".format(self.classname))
        print("  dims:      {}".format(self.dims))
        print("  flags:     {}".format(self.flags))
        print("  version:   {}".format(self.version))
        print("  colormaps: ")
        for m in self.colormaps:
            print(m)

        for i in self.get_images():
            print("    image: {}".format(self.get_image(i).name))
            for l in self.get_image(i).get_layers():
                print("      layer: {}".format(self.get_image(i).get_layer(l).name))

        print("\n")

    def get_image(self, key):
        """ Returns an image given its key. """
        result = False
        if key in self.images:
            result = self.images[key]
        return result

    def get_images(self):
        """ Returns all images. """
        for i in self.images:
            yield i

    def get_image_names(self):
        """ Returns list of image names. """
        return list(self.images.keys())

    def set_parameter_table(self, table):
        """ Set parameter table using a deep copy. """
        self.parametertable = table.copy(deep=True)

    def add_parameter(self, name, type):
        """ Add a parameter to the list of parameters for the CIS. """
        # check for duplicates
        self.parameterlist.append([name, type])

    def add_variable(self, name, type, min, max):
        """ Add a variable to the set of variables. """
        # check for duplicates
        self.variables[name] = {'type':type, 'min':min, 'max':max}

    def add_image(self, name):
        """ Add an image to the set of images in the CIS. """
        # check for duplicates
        self.images[name] = image.image(name)

        return self.images[name]

    def get_variables(self):
        """ Return all variables. """
        for i in self.variables:
            yield i

    def get_variable(self, name):
        """ Return a variable. """

        variable = None
        if name in self.variables:
            variable = self.variables[name]

        return variable

    def get_image(self,name):
        """ Return an image. """
        image = None

        if name in self.images:
            image = self.images[name]

        return image

    def get_colormap(self,name):
        """ Return a colormap. """
        colormap = None

        if name in self.colormaps:
            colormap = self.colormaps[name]

        return colormap

    def add_colormap(self, name, path):
        """ Add a colormap to the set of colormaps. """
        #if colormap not in dict
        if (name not in self.colormaps):
            self.colormaps[name] = colormap.colormap(path)

    def remove_colormap(self, name):
        """ Remove a colormap from the set of colormaps. """
        self.colormaps.pop(name)

    def get_colormaps(self):
        """ Return all colormaps. """
        for i in self.colormaps:
            yield i

    def set_dims(self, w, h):
        """ Set the dimensions of the CIS given a width and height. """  
        self.dims = [w, h]
